%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% classify.pl
% ===========
%
% Klassifikation einer Klausenmenge im TPTP-Format - Hauptdatei
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- [pvdclass].
:- [rename].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% classify(+FileName, -FileName1, -Type)
% ============
% FileName: Dateiname eines Files, das eine Klausenmenge im TPTP-Format enhaelt
%    Falls F das Zeichen "." enthlt, mu F als '...' geschrieben werden!
% FileName1: Name der umbenannten TPTP-Datei
% Type: Klasse der TPTP-Datei (pvd, pvdeq oder unknown)
%
% Klassifiziert eine TPTP-Klausenmenge
% Schreibt ein Renaming in Datei
% Bis jetzt behandelt das Prdikat den Import von anderen TPTP-Dateien noch
% nicht!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

classify(FileName, FileName1, Type) :-
%    writef("Klassifiziere %w:\n", [FileName]),
    read_data(FileName),
    import_data,
    retract_data(TPTPData),
    pvd_classify(TPTPData, Renaming, Type, FileName),
    !,
    write_renaming(Renaming, Type, FileName, FileName1).

classify(FileName, FileName1, unknown) :-
    write_renaming(_, unknown, FileName, FileName1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% read_data(F):
% =============
%
% F: Dateiname
% read_data liest Daten aus einer Datei ein und trgt sie mit assert
% in die Datenbank ein.
% blicherweise wird das Prdikat gemeinsam mit retract_data verwendet,
% das die Daten aus der Datenbank wieder entfernt und in einer Liste
% zusammenfasst.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

read_data(F) :-
    seeing(OldInput),
    see(F),
    repeat,
        read(Line),
        (Line==end_of_file ->
	    true
	    ;
	    assert(Line),
	    fail
	),
    !,
    seen,
    see(OldInput).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% import_data:
% ============
%
% Liest alle Dateien ein, die in include-Anweisungen im TPTP-File
% angefhrt sind
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

import_data :-
	retract(include(FileName)),
        read_data(FileName),
	import_data.

import_data:-
    	\+ retract(include(_)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% retract_data(-ClauseSet):
% =========================
%
% ClauseSet: Liste von Klausen, die aus der Datenbank extrahiert wurden
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

retract_data([]):-
    	\+ input_clause(_, _, _).

retract_data([Clause | Clauses]) :-
	input_clause(X, Y, Clause0),
	retract(input_clause(X, Y, Clause0)),
        convert_cl(Clause0, 0, Clause),
	retract_data(Clauses).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% write_renaming(+Renaming, +Type, +TPTPData, +FileName):
% =============================
% F: Dateiname des TPTP-Files, das analysiert wurde
% PS: Renaming fr die Prdikatensymbole
% Type: Klasse der TPTP-Klausen (pvd, pvdeq oder unknown)
%
% Erzeugt aus F einen neuen Dateinamen (*.res statt *.p) und schreibt
% das Renaming in *.res. Wenn es kein Renaming gibt (also Type = unknown ist),
% wird ein entsprechender Kommentar in die Datei geschrieben.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_renaming([], _Type, FileName, FileName) :-
    writef("%w liegt in einer unbekannten Klasse\n", [FileName]),
    !.
% eher unsauber...


write_renaming(Renaming, pvd, FileName, FileName1):-
    writef("%w liegt in PVD\n", [FileName]),
    extract_renaming_string(FileName, FileName1),
    !,
    write_renaming_1(FileName, FileName1, Renaming),
    transform_clauses(FileName, Renaming, pvd).

write_renaming(Renaming, pvd_geq, FileName, FileName1):-
    writef("%w liegt in PVD_g^eq\n", [FileName]),
    extract_renaming_string(FileName, FileName1),
    !,
    write_renaming_1(FileName, FileName1, Renaming),
    transform_clauses(FileName, Renaming, pvdeq).

write_renaming(_Renaming, unknown, FileName, FileName):-
    writef("%w liegt in einer unbekannten Klasse\n", [FileName]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% write_renaming_1(+F_Old, +F, +Ps):
% ================================
% F_Old: Dateiname des TPTP-Files
% F: Dateiname des Renamings
% Ps: Renaming
% 
% Schreibt einen Kommentar und das Renaming Ps in die Datei mit Namen F.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_renaming_1(F_Old, F, Ps):-
%    writef("Schreibe Renaming in die Datei "), writef(F), nl,
    telling(OldOutput),
    tell(F),
    writef("% Filename:"), writef(F_Old), nl,
    writef("% Renaming:"), nl,
    write(Ps),
    writef("."), nl,
    told,
    tell(OldOutput).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% write_renamed_data(+FileName, +RenamedData, +Type):
% ============================================
% FileName: Dateiname der TPTP-Datei
% RenamedData: Umbenannte Klausenmenge
% Type: pvd oder pvdeq
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

transform_clauses(FileName, Renaming, pvd):-
    extract_pvd_string(FileName, FileName1),
    transform_file(FileName, Renaming, FileName1).

transform_clauses(FileName, Renaming, pvdeq):-
    extract_pvdeq_string(FileName, FileName1),
    transform_file(FileName, Renaming, FileName1).

transform_file(FileName, Renaming, FileName1) :-
    seeing(OldInput),
    see(FileName),
    telling(OldOutput),
    tell(FileName1),
    repeat,
        read(Line),
	(Line==end_of_file ->
	    true
	    ;
	    rename_clause(Line, Renaming, NewLine),
	   numbervars(NewLine, 0, _R),
	   writef("%t.\n", [NewLine]),
	   fail
	),
    !,
    told,
    tell(OldOutput),
    seen,
    see(OldInput).




