%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% remove_tautologies(+Cs, -Ds):
%--------------------------------
% Cs: Klausenmenge
% Ds: Klausenmenge, die aus Cs durch das Entfernen aller Tautologien hervorgeht
%
% Verwendet das Prdikat is_tautology, um festzustellen, ob eine Klause
% eine Tautologie ist.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

remove_tautologies([], []).
remove_tautologies([C|Cs], Ds):-
    is_tautology(C),!,
    remove_tautologies(Cs, Ds).
remove_tautologies([C|Cs], [C|Ds]):-
    remove_tautologies(Cs, Ds).

%------------------------------------------------------------------------------
% is_tautology(+Cs):
% ------------------
% Cs: Klause
%
% Ist wahr genau dann, wenn Cs ein Atom und sein Duales enthlt
% (also "P oder non P").
%------------------------------------------------------------------------------

is_tautology([_]) :- fail.
is_tautology([A|Cs]):-
    A =.. [++, Atom],
    B =.. [--, Atom],
    member(B, Cs), !.
is_tautology([A|Cs]):-
    A =.. [--, Atom],
    B =.. [++, Atom],
    member(B, Cs), !.
is_tautology([_|Cs]):-
    is_tautology(Cs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% remove_unit_clauses(Cs, Ds, M):
% -------------------------------
% Cs: Klausenmenge
% Ds: Klausenmenge, entsteht aus Cs, indem
%     1.) alle Einheitsklausen [L] entfernt werden
%     2.) alle Klausen entfernt werden, die L enthalten
%     3.) L aus allen Klausen entfernt wird, sofern die Klausen dadurch nicht
%         leer werden (falls doch, scheitert das Prdikat!)
% M: Menge aller Literale, die in Einheitsklausen auftauchen.
%
% Verwendet das Prdikat get_units, um alle Einheitsklausen zu ermitteln
% sowie das Prdikat remove_units, um die Klausen zu entfernen.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

remove_unit_clauses(Cs, Ds, Us):-
    get_units(Cs, Us),
    remove_units(Us, Cs, Ds).

%------------------------------------------------------------------------------
% get_units(+Cs, -Ds):
% Cs: Klausenmenge
% Ds: Menge aller Literale, die als Einheitsklause in Cs vorkommen.
%------------------------------------------------------------------------------

get_units([], []).
get_units([[Atom]|Cs], [Atom|Ds]):- !,
    get_units(Cs, Ds).
get_units([_|Cs], Ds):-
    get_units(Cs, Ds).

%------------------------------------------------------------------------------
% remove_units(+Us, Cs, -Ds):
% ---------------------------
% Us: Menge von Literalen, die in Einheitsklausen vorkommen.
% Cs: Klausenmenge
% Ds: Klausenmenge, geht aus Cs hervor (s.o.)
%------------------------------------------------------------------------------

remove_units([], Cs, Cs).
remove_units([L | Ls], Cs0, Ds):-
    remove_units_l(L, Cs0, Cs1),
    remove_units(Ls, Cs1, Ds).

remove_units_l(_, [], []).
remove_units_l(L, [C | Cs], Ds) :-
    member(L, C),
    !,
    remove_units_l(L, Cs, Ds).
remove_units_l(L, [C | _], _) :-
    remove_dual_literal(L, C, []),
    !,
    fail.
remove_units_l(L, [C | Cs], [D | Ds]) :-
    remove_dual_literal(L, C, D),
    remove_units_l(L, Cs, Ds).

%------------------------------------------------------------------------------
% remove_dual_literal
%------------------------------------------------------------------------------
remove_dual_literal(_, [], []).
remove_dual_literal(L, [C | Cs], Ds) :-
    is_dual(L, C),
    !,
    remove_dual_literal(L, Cs, Ds).
remove_dual_literal(L, [C | Cs], [C | Ds]) :-
    remove_dual_literal(L, Cs, Ds).

is_dual(L1, L2) :-
    L1 =.. [++, Atom],
    L2 =.. [--, Atom].
is_dual(L1, L2) :-
    L1 =.. [--, Atom],
    L2 =.. [++, Atom].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% remove_pure_literals(+Cs, -Ds, -M):
% ----------------------------------------
% Cs: Klausenmenge
% Ds: Klausenmenge, entsteht aus Cs, indem alle Klausen, die Literale
% enthalten, welche nur positiv oder nur negativ vorkommen, entfernt werden. 
% M: Alle Literale, die nur positiv oder nur negativ vorkommen
%
% Das Prdikat verwendet classify_lits, um fr jedes Literal festzulegen,
% ob es pur vorkommt. remove_pure_lits entfernt die entsprechenden
% Klausen. pure extrahiert aus der Klassifikation der Literale jene, die 
% pur vorkommen.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

remove_pure_literals(Cs, Ds, M) :-
    classify_lits(Cs, [], Class),
    remove_pure_lits(Cs, Class, Ds),
    pure(Class, M).

%------------------------------------------------------------------------------
% classify_lits(+Cs, +Ds0, -Ds):
% ------------------------------
% Cs: Klausenmenge
% Ds0: Akkumulator
% Ds: Liste mit Elementen der Form sign(PS, +/-/*)
%     + steht fr "Atom kommt nur positiv vor"
%     - steht fr "Atom kommt nur negativ vor"
%     * steht fr "Atom kommt sowohl positiv als auch negativ vor"
%
% Verwendet classify_lits_c, um eine einzelne Klause zu klassifizieren
%------------------------------------------------------------------------------

classify_lits([], Class, Class).
classify_lits([C|Cs], Ds0, Ds):-
    classify_lits_c(C, Ds0, Ds1),
    classify_lits(Cs, Ds1, Ds).

%------------------------------------------------------------------------------
% classify_lits_c(+Cs, +Ds0, +Ds):
% --------------------------------
% Cs: Klause
% Ds0: Akkumulator, enthlt Klassifikation der Atome, soweit sie aus
%      den vorigen Klausen bereits bekannt ist
% Ds: Liste aller vorkommenden Atome mit ihrer Klassifikation (+,-,*)
%------------------------------------------------------------------------------

classify_lits_c([], Class, Class).
classify_lits_c([L | Ls], Cs, Ds):-
    L =.. [Sign, Atom],
    classify_lit(Atom, Sign, Cs, Ds1),
    classify_lits_c(Ls, Ds1, Ds).

%------------------------------------------------------------------------------
% classify_lit(+Atom, +Sign, +Cs, -Ds):
% -------------------------------------
% Atom: ein Atom (Aussagensymbol)
% Sign: Vorzeichen, mit dem das Atom vorkommt
% Cs: Enthlt Klassifikation der Atome, soweit sie bereits bekannt ist
% Ds: Ergnzt Cs um die Klassifikation von Atom (kann neu aufgenommen werden,
%     gleich bleiben oder auf * gendert werden)
%------------------------------------------------------------------------------

classify_lit(Atom, Sign, [sign(Atom, Sign) | Cs], [sign(Atom, Sign) | Cs]) :-
    !.
classify_lit(Atom, _, [sign(Atom, _) | Cs], [sign(Atom, *) | Cs]) :-
    !.
classify_lit(Atom, Sign, [C | Cs], [C | Ds]) :-
    classify_lit(Atom, Sign, Cs, Ds).
classify_lit(Atom, Sign, [], [sign(Atom, Sign)]).

%------------------------------------------------------------------------------
% remove_pure_lits(Cs, Ds, Es):
% Entfernt die Klausen, in denen pure Literale vorkommen.
%------------------------------------------------------------------------------

remove_pure_lits([], _, []).
remove_pure_lits([C|Cs], Ds, Es):-
    contains_pure_lit(C, Ds), !,
    remove_pure_lits(Cs, Ds, Es).
remove_pure_lits([C|Cs], Ds, [C|Es]):-
    remove_pure_lits(Cs, Ds, Es).

contains_pure_lit([L|_], Ds):-
    L =.. [Sign, Atom],
    member(sign(Atom, Sign), Ds).
contains_pure_lit([_|Ls], Ds):-
    contains_pure_lit(Ls, Ds).

pure([], []).
pure([sign(L, +)|Ls], [L1|Cs]):-
    L1 =.. [++, L],
    !,
    pure(Ls, Cs).
pure([sign(L, -)|Ls], [L1|Cs]):-
    L1 =.. [--, L],
    !,
    pure(Ls, Cs).
pure([_|Ls], Cs):-
    pure(Ls, Cs).
    
    
    








