%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pvdclass.pl
% ===========
%
% Klassifikation einer Klausenmenge im TPTP-Format
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- [matrix].
:- [seteq].
:- [filename].
:- op(500, fx, ++).
:- op(500, fx, --).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% is_var(+Term):
% ==============
%
% Term: Beliebiger Prolog-Term
% Liefert true zurueck gdw. Term die Form '$V'(_) hat.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

is_var('$V'(_)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pvd_classify (+ClauseSet, -Renaming, -Type):
% =======================
%
% ClauseSet: Klausenmenge
% Renaming: Liste von Literalen der Form ++PS oder --PS 
% Class: Klasse, in der sich die Klausenmenge befindet.
%
% Falls das Praedikat nicht erfuellbar ist, liegt die Klausenmenge nicht in
% PVD und nicht in PVD_g^=. Ansonsten ist M ein renaming
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pvd_classify(ClauseSet, Renaming, pvd_geq, F) :-
        contains_equation(ClauseSet), !,
	all_equations_are_ground(ClauseSet),
	pvd_constraints(ClauseSet, Ps0, []), 
        extract_dp_string(F, F1),
	remove_equation_symbol(Ps0, Ps1),
        list_to_set(Ps1, Ps2),
        telling(Old), 
        tell(F1),
	    show_formulas(Ps2),
        told, 
        tell(Old),
        !,
        no_contradictory_unit_clauses(Ps2),
	find_model(Ps2, Renaming).

pvd_classify(ClauseSet, Renaming, pvd, F) :-
	pvd_constraints(ClauseSet, Ps0, []),
        extract_dp_string(F, F1),
        list_to_set(Ps0, Ps1),
        telling(Old), 
        tell(F1),
	    show_formulas(Ps1),
        told, 
        tell(Old),
        !,
        no_contradictory_unit_clauses(Ps1),
	find_model(Ps1, Renaming).

show_formulas([]).
show_formulas([C|Cs]) :-
	write(C),
	writef("\n"),
	show_formulas(Cs).


% pvd_constraints(+ClauseSet, +Ps0, -Ps)
% =================================
%
% ClauseSet: Klausenmenge
% PS0: Akkumulator
% PS: Liste von Listen von aussagenlogischen Literalen, die angeben, ob ein
%     Praedikatensymbol umbenannt werden muss oder nicht
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pvd_constraints([], Ps, Ps).
pvd_constraints([Clause | Clauses], Ps0, Ps) :-
	pvd_constraints_4_clause(Clause, Ps0, Ps1),
	pvd_constraints(Clauses, Ps1, Ps).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pvd_constraints_4_clause(+ClauseSet, +Ps0, -Ps)
% ===============================================
%
% ClauseSet: Klausenmenge
% Ps0: Akkumulator
% Ps: Beschreibt renaming fuer ClauseSet
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pvd_constraints_4_clause(C, Ps0, Ps) :-
	pvd_vars(C, _Vs0, Vs1),
	pvd_vars2props(Vs1, Ps0, Ps).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pvd_vars(+Literals, + Vs0, -Vs)
% ===============================
%
% Literals: Liste von Literalen
% Vs0: Akkumulator
% Vs: Liste der in Literals vorkommenden Variablensymbole mit max. Tiefe
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pvd_vars([], Vs, Vs).
pvd_vars([Literal | Literals], Vs0, Vs) :-
	Literal =.. [Sign,Atom],
	functor(Atom, PS, _),
	P =.. [Sign, PS],
	pvd_collect(Atom, 0, P, Vs0, Vs1),
	pvd_vars(Literals, Vs1, Vs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pvd_collect(+Term, +Depth, +Lit, +Vs0, -Vs1)
% ================================================
%
% Term: Ein Term
% Depth: Tiefe, in der man sich derzeit befindet
% Lit: Aussagenlogisches Literal bestehend aus Praedikatensymbol und
%      dazugehoerigem Vorzeichen
% Vs0: Akkumulator
% Vs1: Liste der Variablensymbolen mit maximaler Tiefe 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pvd_collect(T, D, P, Vs0, Vs) :-
	is_var(T), !,
	pvd_insert(T, D, P, Vs0, Vs).
pvd_collect(T, D, P, Vs0, Vs) :-
	T =.. [_|Ts],
	D1 is D+1,
	pvd_collects(Ts, D1, P, Vs0, Vs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pvd_collects(+Terms, +Depth, +Lit, +Vs0, +Vs)
% =============================================
% Terms: Liste von Termen
% Depth: Aktuelle Tiefe
% Lit: Aussagenlogisches Literal
% Vs0: Akkumulator
% Vs: Liste der gefundenen Variablen mit maximaler Tiefe
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pvd_collects([], _D, _P, Vs, Vs).
pvd_collects([T|Ts], D, P, Vs0, Vs) :-
	pvd_collect(T, D, P, Vs0, Vs1),
	pvd_collects(Ts, D, P, Vs1, Vs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pvd_insert(+Var, +Depth, +Lit, +Vs0, -Vs)
% =========================================
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pvd_insert(V, D, P, Vs, Vs) :-
	pvd_insert_nc(V, D, P, Vs), !.
pvd_insert(V, D, P, Vs0, Vs) :-
	pvd_insert_c(V, D, P, Vs0, Vs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pvd_insert_nc(V1, D1, P1, Vs)
% =============================
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
pvd_insert_nc(V1, D1, P1, [v(V1, D1, [P1|_])|_]) :- !.
pvd_insert_nc(V1, D1, P1, Vs0) :-
	Vs0 = [v(V2,D2,Ps2)|Vs1],
	pvd_insert_nc1(V1, V2, D1, D2, P1, Ps2, Vs1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
pvd_insert_nc1(V, V, D1, D2, P1, Ps2, _Vs) :- !,
	pvd_insert_nc2(D1, D2, P1, Ps2).
pvd_insert_nc1(V1, _V2, D1, _D2, P1, _Ps2, Vs) :-
	pvd_insert_nc(V1, D1, P1, Vs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
pvd_insert_nc2(D, D, P, Ps) :- !,
	member(P, Ps).
pvd_insert_nc2(D1, D2, _P, _Ps) :-
	D1 < D2.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
pvd_insert_c(V, D, P, [v(V,_,_)|Vs], [v(V,D,[P|_])|Vs]) :-
	 !.
pvd_insert_c(V, D, P, [VE|Vs0], [VE|Vs]) :-
	pvd_insert_c(V, D, P, Vs0, Vs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pvd_vars2props(Vs, Ps, Ps) :-
	var(Vs), !.
pvd_vars2props([v(_,_,P)|Vs], [P|Ps0], Ps) :-
	append(P, [], P),
	pvd_vars2props(Vs, Ps0, Ps).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% contains_equation(+ClauseSet):
% ===============================
%
% Cs: Menge von Klausen
% 
% true gdw. Cs Gleichungen der Form ++equal(...) oder --equal(...) enthaelt.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

contains_equation([]) :- 
	fail, 
	!.
contains_equation([Clause | _Clauses]):-
	contains_equation_cl(Clause).
contains_equation([_Clause | Clauses]):-
	contains_equation(Clauses).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% contains_equation_cl(+Clause)
% ====================================
%
% Clause: Menge von Literalen
%
% true gdw. ein Literal die Form ++equal(...) oder --equal(...) hat 
% ist.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

contains_equation_cl([]):-
	fail,
	!.
contains_equation_cl([Literal | _Literals]):-
	is_equation(Literal).
contains_equation_cl([_Literal | Literals]):-
	contains_equation_cl(Literals).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% is_equation(+Literal)
% ============================
% 
% Literal: Ein Term der Form ++PS(...) oder --PS(...)
%
% true gdw. das Praedikatensymbol "eq" ist und keine Variablen vorkommen
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

is_equation(Literal) :-
	Literal =.. [_Sign | Atom],
        Atom = [Atom1],
	Atom1 =.. [equal | _Arguments].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ground(+Arguments)
% ==================
%
% Arguments: Liste von Termen
%
% true gdw. wenn kein Argument eine Variable enthaelt
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

is_ground([]).
is_ground([Term | Terms]):-
	ground_term(Term),
	is_ground(Terms).
%        , !,
%        writef("***** GRUND\n").

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ground_term(Term):
% ==================
%
% Term: Ein Term
%
% True gdw. Term weder eine Variable ist noch eine beinhaltet
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ground_term(Term):-
	is_var(Term),   % keine Variable
        !,
	fail.

ground_term(Term) :-
	atomic(Term),   % Konstante
	!.

ground_term(Term):-
	Term =.. [_FS | Args], % Funktion
	ground(Args).          % Alle Argumente ground

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
all_equations_are_ground([]).
all_equations_are_ground([Clause | Clauses]):-
    all_equations_are_ground_cl(Clause),
    all_equations_are_ground(Clauses).

all_equations_are_ground_cl([]).
all_equations_are_ground_cl([Literal | Literals]) :-
    is_equation(Literal), !,
%    writef("\tTeste "),
%    write(Literal),
%    writef("\n"),
    Literal =..[_Sign | Atom],
    Atom = [Atom1],
    Atom1 =.. [equal | Arguments],
    is_ground(Arguments), !,
    all_equations_are_ground_cl(Literals).
all_equations_are_ground_cl([_Literal | Literals]):-
    all_equations_are_ground_cl(Literals).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% remove_equation_symbol(+Cs, -Ds):
% ===========================
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
remove_equation_symbol([],[]).
remove_equation_symbol([C|Cs],[D|Ds]) :-
	remove_equation_symbol2(C,D),
	remove_equation_symbol(Cs,Ds).
	

remove_equation_symbol2([], []).

remove_equation_symbol2([C | Cs], Ds):-
	C =.. [_S | [equal]],
	!,
	remove_equation_symbol2(Cs, Ds).

remove_equation_symbol2([C | Cs], [C|Ds]):-
	remove_equation_symbol2(Cs, Ds).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% no_contradictory_unit_clauses(ClauseSet):
% ==================================
%
% ClauseSet: Klausenmenge
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

no_contradictory_unit_clauses([]).
no_contradictory_unit_clauses([[Ps] | ClauseSet]) :-
    Ps =.. [++, Atom],
    Ps2 =.. [--, Atom],
    \+ member([Ps2], ClauseSet),
    no_contradictory_unit_clauses(ClauseSet).
no_contradictory_unit_clauses([[Ps] | ClauseSet]) :-
    Ps =.. [--, Atom],
    Ps2 =.. [++, Atom],
    \+ member([Ps2], ClauseSet),
    no_contradictory_unit_clauses(ClauseSet).
no_contradictory_unit_clauses([Clause | ClauseSet]) :-
    length(Clause, L),
    L > 1,
    no_contradictory_unit_clauses(ClauseSet).













