:- op(500, fx, ++).
:- op(500, fx, --).

my_member(E, [E|_]).
my_member(E, [_|L]) :-
	my_member(E,L).

%------------------------------------------------------------------------------
% compute_setting(+TPTPfile, SettingFile)
% -----------------------------------
%    TPTPfile: Filename of a file with tptp-clauses
%    Setting : Filename for result
%
%    reads all tptp-clauses from the file
%    with filename TPTPfile.
%    asserts these clauses and calls
%    find_setting/0. The result of find_setting is written to SettingFile
%------------------------------------------------------------------------------
compute_setting(TPTPfile, SettingFile) :-

	seeing(OldInput),
	see(TPTPfile),
	repeat,
	   read(Clause),
 	   (Clause == end_of_file ->
	    true
	   ;
	    assert_clause(Clause), % Save only relevant part of clause
 				   % to the database
	fail
	),
	!,
	seen,
	see(OldInput),
        telling(OldOutput),
        tell(SettingFile),
	find_setting,
        told,
        tell(OldOutput).

%------------------------------------------------------------------------------
% assert_clause(TPTPClauseSet):
%-----------------------
% TPTPClauseSet: tptp-clause with syntax input_clause(Name, Type, ClauseSet), 
% where ClauseSet is given by [Clause1, ..., ClauseN] with 
% ClauseI= [Literal1, ..., LiteralM]. 
% Each literal has the syntax ++p(t1,...tR) or --p(t1,...,tR).
%
% This predicate asserts the clause input_clause(Contents1), where Contents1
% is derived from Contents by replacing each variable by $V(N), N ranging from
% 1 to the number of variables.
%------------------------------------------------------------------------------
assert_clause(input_clause(_Name, _Type, Clause)) :-
	convert_cl(Clause, 0, NewClause), 
        !,
	assertz(input_clause(NewClause)).

%------------------------------------------------------------------------------
% find_setting(?M)
% ----------------
%   M: Setting for the clauses in the database
%      which obeys pvd
%
%   This predicate operates on the database which is
%   formed by predicates "input_clause([L1, ..., Ln])". 
%
%   After reading a clause, the coresponding element is 
%   removed from the database. After the clauses are 
%   read, the predicate evaluates m_pvd with the set of 
%   these clauses and returns  the resulting representation 
%   of a PVD-setting for S.
%------------------------------------------------------------------------------
find_setting :-
	findall(Clause, retract(input_clause(Clause)), S), !,
	m_pvd(S, M, Type),
        write(Type), nl,
        write(M).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%------------------------------------------------------------------------------
% c_s(+ClauseSet, -Setting):
% --------------------------
% ClauseSet: List of clauses (lists of literals
% of the form ++ Atom or -- Atom)
% Setting: List of Literals of the form
% ++ p(X1, ..., Xn) or --p(X1, ..., Xn), where
% n is the arity of p.
%
% principal predicate for finding a setting
% for a clause set interactively
% 
% Converts the clause set to $V-Form, then
% applies m_pvd to find a setting
%------------------------------------------------------------------------------

c_s(ClauseSet, Setting, Type) :-
	convert_cl(ClauseSet, 0, NewClauseSet),
	!,
	m_pvd(NewClauseSet, Setting, Type).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%------------------------------------------------------------------------------
% m_pvd(+S, ?M, ?Type)
% -------------
%   S: Set of clauses in the syntax of tptp, without "input_clause()"
%   M: Setting for S which obeys pvd
%   Type: Type of clauses: pvd or pvdgeq
%
%   This is the principal predicate for finding a PVD-setting
%   True when M is the represetation of a PVD-setting for S
%------------------------------------------------------------------------------

m_pvd(S, M, Type) :-
    prepare_clauses(S, CS, _, Type), 
    !,
    pvd(CS, M).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%------------------------------------------------------------------------------
% convert_cl(Cl, N, Cl1)
% ------------------
%   Cl: clause
%   N : number of the next free variable
%   Cl1 : Output, where each variable is replaced by $V(N)
%
%   This predicate converts a clause to
%   an equivalent clause where all 
%   variables are replaced by $V(N),
%   with N ranging between 0 and the
%   number of variables.
%------------------------------------------------------------------------------

convert_cl([], _, []).
convert_cl([Lit | Lits], N, Clause) :-
    convert_lit(Lit, N, M, Lit1),
    convert_cl(Lits, M, Lits1),
    Clause = [Lit1 | Lits1].

%------------------------------------------------------------------------------
% convert_lit(+Lit, N, M, -Lit1)
% -----------------------
% Lit: A literal
% N  : the numbers of the next variable
% M  : the number of the butnext variable
% Lit1: The literal computed from Lit by replacing each variable by $V(N)
%
%   converts a literal such that each variable
%   is replaced by '$V'(N). N is incremented with each variable.
%------------------------------------------------------------------------------

convert_lit('$V'(N), N, M, '$V'(N)) :-
    !,
    M is N+1. 
convert_lit(Term, N, M, Term1) :-
    Term =.. [F | Args],
    convert_arguments(Args, N, M, Args1),
    Term1 =.. [F | Args1].

%------------------------------------------------------------------------------
% convert_arguments(+Args, +N, -M, -Args1):
% -----------------------------------------
% Args  : List of terms
% N     : Number of next variable
% M     : Number of next variable after converting all arguments
% Args1 : List of converted terms
%------------------------------------------------------------------------------
convert_arguments([], N, N, []).
convert_arguments([Arg | Args], N, M, [Arg1 | Args1]) :-
    convert_lit(Arg, N, M1, Arg1),
    convert_arguments(Args, M1, M, Args1).

%------------------------------------------------------------------------------
% prepare_clauses(+S, -CS, -PSs, -Type)
% ------------------------------
%   True when PSs is an ordset of all predicate symbols occuring
%   in the set of clauses S, and CS is the conversion of S,
%   where every Literal in every clause is changed to
%   Literal-[Variable-MaxDepth|_].
%   Type is type of clause set: pvd or pvdgeq
%   [tested 1/11/91]
%------------------------------------------------------------------------------

prepare_clauses(S, CS, PSs, Type) :-
    prepare_clauses_dl(S, CS, [], PSdups, [], Type),
    sort(PSdups, PSs).

%------------------------------------------------------------------------------
% prepare_clauses_dl(ClauseSet, Conversion0, Conversion1, Preds0, Preds1, Type):
% ------------------------------------------------------------------------
% Auxiliary predicate for preparing clauses
%------------------------------------------------------------------------------

prepare_clauses_dl([], CS, CS, PSs, PSs, pvd).
prepare_clauses_dl([Clause|Clauses], [NewCl|CS0], CS, PSs0, PSs, Type) :-
    prepare_clause_cl(Clause, NewCl, [], PSs0, PSs1, Type1),
    prepare_clauses_dl(Clauses, CS0, CS, PSs1, PSs, Type2),
    ( 
      (Type1 = pvdgeq 
       ; 
       Type2 = pvdgeq)
      -> 
         Type = pvdgeq
      ;
         Type = pvd).

%------------------------------------------------------------------------------
% prepare_clause_cl(+Clause, -ConvCl0, -ConvCl, -PSs0, -PSs, Type)
% ----------------------------------------------------------
%   True when ConvCl0\ConvCl is a difflist of converted
%   literals of Clause, and PSs0\PSs is a difflist of
%   predicate symbols occuring in Clause.
%   [tested 23/10/91]
%------------------------------------------------------------------------------

prepare_clause_cl([], Cl, Cl, Ps, Ps, pvd).
prepare_clause_cl([Lit|Lits], [Lit-[]|Cl0], Cl, 
    PSs0, PSs, pvdgeq) :-
    (   Lit = ++ Atom;   
        Lit = -- Atom
    ),
    %AB: If the clause is an equation, do not
    %    add the equality predicate symbol
    functor(Atom, eq, 2),
    make_varlist(Atom, Vars),
    Vars = [],
    prepare_clause_cl(Lits, Cl0, Cl, PSs0, PSs, _).
prepare_clause_cl([Lit|Lits], [Lit-Vars|Cl0], Cl, 
                  [PS/A|PSs0], PSs, Type) :-
    (   Lit = ++ Atom;   
        Lit = -- Atom
    ),
    functor(Atom, PS, A),
    make_varlist(Atom, Vars),
    prepare_clause_cl(Lits, Cl0, Cl, PSs0, PSs, Type).

%------------------------------------------------------------------------------
% make_varlist(+Atom, -Vars)
% --------------------------
%   True if Vars is an ordset of pairs -/2 of variables and
%   their respective maximal depths of occurence in Atom.
%   [tested 23/10/91]
%------------------------------------------------------------------------------

make_varlist(Atom, Vars) :-
    setof(  depth(X,MDepth), mdepth(Atom, X, MDepth), Vars),
    !. % if Atom contains Vars, but then only once.
make_varlist(_, []). % if Atom is ground

    mdepth(Atom, X, MDepth) :-
        var_occ(Atom, X),
        tmax_t_t(X, Atom, MDepth).

% is_setting(+PredSymbols, -Setting)
% ----------------------------------
%   True if Setting is the representation of a setting.
%   [tested 23/6/97]

is_setting([], []).
is_setting([PS/A|PSs], [Lit|Lits]) :-
    functor(Atom, PS, A), !,
    signed(Atom, Lit),
    is_setting(PSs, Lits).

% signed(+Atom, -Literal)
% =======================

signed(A, ++A).
signed(A, --A).


% pvd(+S, +M)
% -----------
%   True, when M is a PVD-setting for S

pvd([], _).
pvd([C|Cs], M) :-
    cl_analysis(C, M, _, _, Vneg, Vpos),
%    write('Negative Variables:'), write(Vneg), nl,
%    write('Positive Variables:'), write(Vpos), nl,
    !, % no need to backtrack--once a clause fails, pvd fails
    pvd_cl(Vneg, Vpos),
%    write('The clause'), write(C), write('obeys pvd'), nl,
    pvd(Cs, M).


% pvd_cl(+Vneg, +Vpos)
% ------------------------------------------
%   True when PVD1-2 hold for clauses and settings with 
%   Vneg and Vpos, the variables-maxdepths occuring in
%   Cneg and Cpos, respectively.

pvd_cl(_, []).
pvd_cl(Vneg, [Var-MDpos | Vpos]) :-
    my_member(Var-MDneg, Vneg),
%    write('Variable '), write(Var), write(':'),  nl,
%    write('Max. neg. Depth:'), write(MDneg), nl,
%    write('Max. pos. Depth:'), write(MDpos), nl,
    MDpos =< MDneg,
    !, 
    pvd_cl(Vneg, Vpos).

% cl_analysis(+C, +M, -Cneg, -Cpos, -Vneg, -Vpos)
% -----------------------------------------------
%   Returns for given clause C and setting M the negative and
%   positive parts of C w.r.t. M in Cneg and Cpos, respectively,
%   and a list of variables with their respective maximum
%   term depths in Cneg and Cpos as Vneg and Vpos.
%   [tested 1/11/91]

cl_analysis(C, M, Cneg, Cpos, Vneg, Vpos) :-
    cl_analysis1(C, M, Cneg, [], Cpos, [], Vn, [], Vp, []),
    max_var_depth(Vn, Vneg),
    max_var_depth(Vp, Vpos).

cl_analysis1([], _, Cn, Cn, Cp, Cp, Vn, Vn, Vp, Vp).
cl_analysis1([L|Ls], M, Cn1, Cn, [L|Cp1], Cp,
            Vn1, Vn, Vp0, Vp) :-
    is_true(L, M),
    !,
    collect_vars(L, Vp0, Vp1),
    cl_analysis1(Ls, M, Cn1, Cn, Cp1, Cp, Vn1, Vn, Vp1, Vp).
cl_analysis1([L|Ls], M, [L|Cn1], Cn, Cp1, Cp,
            Vn0, Vn, Vp1, Vp) :-
    collect_vars(L, Vn0, Vn1),
    cl_analysis1(Ls, M, Cn1, Cn, Cp1, Cp, Vn1, Vn, Vp1, Vp).


% is_true(+Literal, +Setting)
% ===========================
%AB: If the literal is ++eq(_,_) or --eq(_,_), then
% it is automatically true.
is_true(L-_, _) :-
    (L = ++A
    ;
    L = --A),
    functor(A, eq, _), 
    !.
%AB: Otherwise, the literal is true iff it is contained
% in the setting.
is_true(L-_, M) :-
    copy_term(M, Setting),
    my_member(L, Setting).


% collect_vars(+L, -V0, -V)
% -------------------------
%   Returns the variables in L with their respective depths
%   in the difference list V0, V as pairs Var-Depth.

collect_vars(_-Vars, V0, V) :-
    collect_vars1(Vars, V0, V).

collect_vars1([], V, V).
collect_vars1([VD|VDs], [VD|V0], V) :-
    collect_vars1(VDs, V0, V).


% max_var_depth(+VarDepths, -MaxVarDepths)
% ----------------------------------------
%   True when MaxVarDepths is an ordlist containing, for 
%   each variable in VarDepths, the one (Var-Depth)-pair
%   where Depth is maximal in VarDepths; uses max_var_depth/4
%   [tested 1/11/91]

max_var_depth(VDs, MVDs) :-
    max_var_depth(VDs, VDs, MVDs, []).


% max_var_depth(Vars, VDs, MVDs0, MVDs)
% -------------------------------------
%    Vars is the list of -/2-pairs to process
%    VDs is a list of -/2-pairs to compare with, i.e. to test if the
%    current Var has the maximum Depth or not. If Var has already
%    been inserted into MVDs, it contains the maximum d/2-entry
%    for Var, prefixed by `-'.
%    MVDs0, MVDs is a difference list of the maximum d/2-entries.
%    [tested 1/11/91]

max_var_depth([], _, MVDs, MVDs).
max_var_depth([Var|Vars], VDs, [Var|MVDs0], MVDs) :-
    is_max_depth(Var, VDs, NewVDs, []),
    !,
    max_var_depth(Vars, NewVDs, MVDs0, MVDs).
max_var_depth([_|Vars], VDs, MVDs0, MVDs) :-
    max_var_depth(Vars, VDs, MVDs0, MVDs).


% is_max_depth(+Var, +VDs, -NewVDs0, +NewVDs)
% --------------------------------------------
%   tests, if Var (a -/2-pair) has the maximum Depth entry in
%   VDs, and if so, returns a new list of variable depths NewVDs that
%   contains all entries of Vars save those for the variable in Var
%   plus a `-'-prefixed -/2-pair marking that the maximum for Var
%   has already been found.
%       It does this by scanning Vars from left to right, until one
%   of the following holds:
%       (i)    all of Vars has been scanned, then Var is the maximum
%       (ii)   there is a `-'-prefixed pair with matching Var, then
%              the maximum has already been found earlier in max_depth
%   [tested 1/11/91]

% (i)
is_max_depth(Var, [], [- Var|NewVDs], NewVDs).

% (ii)
is_max_depth(Var-_, [-(Var-_)|_], _, _) :-
    !,
    fail.

% Var-MDepth could still be the max
is_max_depth(Var-MDepth, [Var-Depth|Vars], NewVDs0, NewVDs) :-
    !,
    MDepth >= Depth,
    is_max_depth(Var-MDepth, Vars, NewVDs0, NewVDs).

% different Vars, so skip it
is_max_depth(V, [Var|Vars], [Var|NewVDs0], NewVDs) :-
    is_max_depth(V, Vars, NewVDs0, NewVDs).



% var_occ(+Term, ?Var)
% --------------------
%   True, if Var is a variable '$V'(_) occuring in Term
%   [tested 23/10/91]

var_occ('$V'(N), '$V'(N)).
var_occ(Term, Var) :-
    functor(Term, _, A),
    var_occ(A, Term, Var).
var_occ(0, _, _) :- !, fail.
var_occ(I, Term, Var) :-
    arg(I, Term, SubTerm),
    (   var_occ(SubTerm, Var)
    ;   J is I-1,
        var_occ(J, Term, Var)).


% tmax_t_t(+Term1, +Term2, ?TMax)
% --------------------------------
%   True, if TMax is the maximum depth of Term1 in Term2
%   [tested 19/10/91]

tmax_t_t(T, T, 0) :- !.
tmax_t_t(_, '$V'(_), -1) :- !.
tmax_t_t(_, PAtom, -1) :-
    atom(PAtom), !.
tmax_t_t(T1, T2, Tmax) :-
    functor(T2, _, Arity),
    tmax_t_t(Arity, T1, T2, -1, SubTmax),
    (SubTmax =:= (-1) -> Tmax is -1; Tmax is SubTmax + 1).

    % tmax_t_t/5
    % ---------
    %	iteratively computes tmax for each SubTerm and returns
    %	the maximum depth of occurence.
    
    tmax_t_t(0,_,_,Tmax, Tmax).
    tmax_t_t(0,_,_,Tmax, Tmax) :- !, fail.
    tmax_t_t(I, T1, T2, SoFar, Tmax) :-
        arg(I, T2, SubTerm),
        tmax_t_t(T1, SubTerm, SubTmax),
        SubTmax > SoFar,
        !,
        J is I-1,
        tmax_t_t(J, T1, T2, SubTmax, Tmax).
    tmax_t_t(I, T1, T2, SoFar, Tmax) :-
        !,
        J is I-1,
        tmax_t_t(J, T1, T2, SoFar, Tmax).
 



