% cl_eval.pl
%

:- module(cl_eval,[is_equal/1,
	is_atom/1,
	is_literal/1,
	is_clause/1,
	is_neq_literals/1,
	is_geq_literals/1,
	is_negative/1,
	is_positive/1,
	new_constant/2,
	compSize/2,
	compSizeProblem/2,
	narrowing/3,
	narrowTerm/4,
	eval/2,
	eval/3]).

:- assertz(user:file_search_path(basic,'.')).
:- assertz(user:file_search_path(trs,'.')).
:- assertz(user:file_search_path(cl,'.')).

:- use_module(library(lists)).
:- use_module(basic(general)).                  % general predicates
:- use_module(basic(term)).

:- use_module(trs(mi_compl),[startKB/3,startKB_problem/3]).
:- use_module(trs(mi_cps),[rewrite/4]).

:- use_module(cl(tptp_parser),[read_arm_as_list/2,mb_parser/2]).

%--------------------------------------------------------------%
%                                                              %
%         KB completion                                        %
%                                                              %
%--------------------------------------------------------------%

% compSize(+Equations, -Rules) <-----
% compSizeProblem(+Problem, -Rules) <---
%
%              these predicates ensure that Rules is a complete
%              set of rewrite rules for equations
%
%              Problem is a name for a certain set of equations
%              listed in mi_data.pl    

compSize(Equations, Rules) :-
    startKB(Equations,sizeOrd,Rules).

compSizeProblem(Problem, Rules) :-
    startKB_problem(Problem,sizeOrd,Rules).


%--------------------------------------------------------------%
%                                                              %
%   definitions                                                %
%                                                              %
%--------------------------------------------------------------%

:- op(500,fx,++).
:- op(500,fx,--).

sign(++).
sign(--).


is_equal(equal(S,T)) :-
	is_term(S),
	is_term(T).

is_equal(equal(f(a),a)).

is_atom(Atom) :-
	   nonvar(Atom),
           Atom =.. [P|PArgs],
	   atomic(P),
           nonvar(PArgs),
           is_termlist(PArgs).

is_atoms([]).
is_atoms([A|As]) :-
	   is_atom(A),
	   is_atoms(As).

is_literal(Lit) :-
	   nonvar(Lit),
	   Lit =.. [S|[At]],
	   sign(S),
           At =.. [P|PArgs],
	   atomic(P),
           is_termlist(PArgs).

:- is_literal(--'L'(X,Y)).
:- is_literal(++love(X)).

is_neq_literal(Lit) :-
	nonvar(Lit),
	Lit =.. [S|[At]],
	sign(S),
        At =.. [P|PArgs],
	atomic(P),	
	\+ (P = 'equal'),
	is_termlist(PArgs).

is_geq_literal(Lit) :-
	ground(Lit),
	nonvar(Lit),
	Lit =.. ['++'|[At]],
	is_equal(At).
	
is_clause([]).
is_clause([L|Ls]) :-
	   is_literal(L),
	   is_clause(Ls).

:- is_clause([++gaze(a)]).
:- is_clause([--love(X)]).
:- is_clause([--love(X),++gaze(X)]).
:- is_clause([--'Q'(f(f(a))),--'Q'(f(f(f(a))))]).

is_clauses([]).
is_clauses([C|Cs]) :-
           is_clause(C),
           is_clauses(Cs).

:- is_clauses([[++gaze(a)],[++love(a)]]).

%
%	is_neq_literals					
%

is_neq_literals([]).
is_neq_literals([L|Ls]) :-
	is_neq_literal(L),
	is_neq_literals(Ls).

:- is_neq_literals([++gaze(a),--love(b)]).
:- \+ is_neq_literals([++gaze(a),--love(b),++equal(a,b)]).

%
%	is_eq_literals					
%

is_geq_literals([]).
is_geq_literals([L|Ls]) :-
	is_geq_literal(L),
	is_geq_literals(Ls).
	
:- is_geq_literals([++equal(sokrates,ariel),++equal(b,c)]).
:- \+ is_geq_literals([--equal(a,b)]).

%--------------------------------------------------------------%
%                                                              %
%   make eval state                                            %
%                                                              %
%--------------------------------------------------------------%


is_eval_state(eval(Atoms, Rules, _HUniverse,_TAs)) :-
	   is_atoms(Atoms),
	   is_rules(Rules).

%
% make_evalState/2 <--
%

make_evalState(Lits,Equations,eval(Ats,Rules,HUniverse,TAs)) :-
	   extract_fs(Lits,HU0s),
	   extract_fs(Equations,HU1s),
	   literals_atoms(Lits,Ats0),
	   literals_atoms(Equations,Eq0s),
	   merge(HU0s,HU1s,HUniverse),
	   atoms_equations(Eq0s,Eqs),
	   compSize(Eqs,Rules),
 	   narrowing(Ats0,Rules,Ats),
	   build_tas(Ats,TAs).

eval_atoms(eval(Ats,_Rs,_HUs,_TAs),Ats).
eval_rules(eval(_Ats,Rs,_HUs,_TAs),Rs).
eval_huniverse(eval(_Ats,_Rs,HUs,_TAs),HUs).
eval_tas(eval(_Ats,_Rs,_HUs,TAs),TAs).

%
% is_negative(+L) <---
%		   ensures that L is negated, provided
%                  L is TPTP syntax
%
% is_positive(+L) <---
%                  repectively

is_negative(L) :-
	   nonvar(L),
           L =.. [--|_At].

is_positive(L) :-
           nonvar(L),
           L =.. [++|_At].
	   
:- is_positive(++'L'(X)).
:- is_negative(--'L'(X,Y)).


%
% narrowing/3 <--
%

narrowing([],_Rules,[]).
narrowing([At0|Ats0],Rules,[At|Ats]) :-
	At0 =.. [P|Args0],
	narrowArgs(Args0,Rules,Args),
	At =.. [P|Args],
	narrowing(Ats0,Rules,Ats).

narrowAtom(Atom0,Rules,Atom) :-
	Atom0 =.. [P|Args0],
	narrowArgs(Args0,Rules,Args),
	Atom =.. [P|Args].

narrowArgs([],_Rules,[]).
narrowArgs([T0|T0s],Rules,[T|Ts]) :-
	narrowTerm(T0,Rules,T),
	narrowArgs(T0s,Rules,Ts).

% the trs has to be confluent 
% for these predicates!

%narrowTerm(_Rules,T,Allrules,T) :-
%	\+ reducible_all(T,Allrules),
%	!.
%narrowTerm(Rules,T0,Allrules,T) :-
%        select(Rule-_Index,Rules,Rules0),
%	narrowTerm2(T0,Rule,T1),
%	narrowTerm(Rules0,T1,Allrules,T).

narrowTerm(T,Rules,T) :-
	\+ reducible_all(T,Rules),
	!.
narrowTerm(T0,Rules,T) :-
	member(Rule-_Index,Rules),
	reduce(T0,Rule,T1),
	narrowTerm(T1,Rules,T).

%narrowTerm2(T0,Rule,T) :-
%	reduce(T0,Rule,T),
%	\+ reducible(T,Rule),
%	!.

reduce(T0,Rule,T) :-
	   rewrite(T0,Rule,_Pos,T),
	   \+ reducible(T,Rule).
reduce(T0,Rule,T) :-
	   rewrite(T0,Rule,_Pos,T1),
	   reduce(T1,Rule,T).

%---------------------------------------------
% reducible(T,Rule) <--
%          true if T is reducible via Rule
%---------------------------------------------

reducible(T,Rule) :-
	rewrite(T,Rule,_,_).

%---------------------------------------------
% reducible_all(T,Rules) <--
%          true if T is reducible by an Rule in
%          Rules
%---------------------------------------------

reducible_all(T,Rules) :-
	member(Rule-_Ind,Rules),
	reducible(T,Rule).

% tests
:- narrowTerm(f(a),[(f(a)=>a)-1],R), R=a.

:- narrowTerm(f(g(c)),[(c=>c_4)-1,(f(c_4)=>c_4)-2,(g(c_4)=>c_4)-3],R), R=c_4.

%:- narrowTerm(f(g(a)),[(c=>c_4)-1,(f(c_4)=>c_4)-2,(g(a)=>c_4)-4,(g(b)=>c_4)-5,(g(c_4)=>c_4)-3],R), R=c_4.

%:- narrowArgs([f(g(a)),f(g(b)),f(g(c))],[(c=>c_4)-1,(f(c_4)=>c_4)-2,(f(f(a))=>c_4)-6,(f(f(b))=>c_4)-7,(f(g(f(a)))=>c_4)-8,(f(g(f(b)))=>c_4)-9,(g(a)=>c_4)-4,(g(b)=>c_4)-5,(g(c_4)=>c_4)-3,(g(g(f(a)))=>c_4)-10,(g(g(f(b)))=>c_4)-11],N), N=[c_4,c_4,c_4].


%
% atoms_equations(Ats,Eqs) <--
%

atoms_equations([],[]).
atoms_equations([equal(S,T)|Ats],[S=T|Eqs]) :-
	atoms_equations(Ats,Eqs).

%
% literals_atoms(+Lits,-Ats) <---
%                  ensures that Ats is a list of
%                  atoms is the atom(Lit), Lit in Lits

literal_atom(Lit,At) :-
	   Lit =..[++|At0],
	   [At] = At0.
literal_atom(Lit,At) :-
	   Lit =..[--|At0],
           [At] = At0.

:- literal_atom(--'L'(X,Y),At).

literals_atoms([],[]).
literals_atoms([L|Lits],[A|Ats]) :-
	literal_atom(L,A),
	literals_atoms(Lits,Ats).


%--------------------------------------------------------------%
%                                                              %
%     evaluation of ground clauses                             %
%                                                              %
%--------------------------------------------------------------%

%
% literalG_true <-----
%

literalG_true(State,Literal) :-
	   literal_atom(Literal,Atom),
           atomG_true(State,Atom).

literalG_false(State,Literal) :-
	\+ literalG_true(State,Literal).

atomG_true(State,Atom) :-
	eval_atoms(State,Atoms),
	eval_rules(State,Rules),
	Atom =.. [P|Args0],
	narrowArgs(Args0,Rules,Args),
	!,
	atomG_true2(P,Args,Atoms).

atomG_true2(equal,[S,S],_Atoms) :-
	!.
atomG_true2(P,Args,Atoms) :-
	Atom =.. [P|Args],
	member(Atom,Atoms).


clauseG_true(State,Clause) :-
	   \+ clauseG_false(State,Clause).

clauseG_false(_State,[]).
clauseG_false(State,[L|Cs]) :-
	   is_negative(L),
           literal_atom(L,A),
           atomG_true(State,A),
	   clauseG_false(State,Cs).
clauseG_false(State,[L|Cs]) :-
	   is_positive(L),
           literal_atom(L,A),
           \+ atomG_true(State,A),
           clauseG_false(State,Cs).


%
% evalState_ground
%

evalState_ground(State,C) :-
	   is_literal(C),
           literalG_true(State,C).
evalState_ground(State,C) :-
           is_clause(C),
           clauseG_true(State,C).
	   


%--------------------------------------------------------------%
%                                                              %
%    non ground clauses                                        %
%                                                              %
%--------------------------------------------------------------%

% extract_fs <--
%

extract_fs([],[]).  
extract_fs([L|Clause],FSs) :-
	is_literal(L),
	extract_literal(L,FS0s),
	extract_fs(Clause,FS1s),
	merge(FS0s,FS1s,FSs).

extract_literal(L,FSs) :-
	literal_atom(L,A),
	extract_atom(A,FSs).

extract_atom(A,FSs) :-
	A =.. [_P|Args],
	extract_pred_list(Args,FSs).

extract_pred_list([],[]).
extract_pred_list([V|Args],FSs) :-
	var(V),
	extract_pred_list(Args,FSs).
extract_pred_list([T|Args],FSs) :-
	nonvar(T),
	extract_term(T,FS0s),
	extract_pred_list(Args,FS1s),
	merge(FS0s,FS1s,FSs).

extract_term(V,[]) :-
	var(V).
extract_term(T,FSs) :-
	nonvar(T),
	T =.. [F|TArgs],
	extract_termlist(TArgs,FS0s,A),
	merge([F/A],FS0s,FSs).

extract_termlist([],[],0).
extract_termlist([V|Args],FSs,N) :-
	var(V),
	extract_termlist(Args,FSs,N0),
	N is N0 + 1.
extract_termlist([T|Ts],FSs,N) :-
	nonvar(T),
	extract_term(T,FS0s),
	extract_termlist(Ts,FS1s,N0),
	merge(FS0s,FS1s,FSs),
	N is N0 + 1.

:- extract_fs([++love(a)],FS).
:- extract_term(a,FS).
:- extract_termlist([a,b],FS,A).
:- extract_literal(++love(a),FS).
:- extract_fs([++'Q'(f(f(X))),--'P'(X,Y)],H).

%
% build_tas(Atoms,TAs) <---
%         generaetes a list of all subterms of Atoms					
%

build_tas([],[]) :-
	!.
build_tas(Ats,TAs) :-
	setof(Sub,atoms_subterm(Ats,Sub),TAs).

atoms_subterm(Ats,Sub) :-
	member(At,Ats),
	subterm(At,Pos,Sub),
	\+ (Pos = []).

:- build_tas([q(a),r(f(b))],X).

%
% new_term/3 <--
%

new_term(F/A,TAs,NT) :-
	new_term(A,TAs,[],NTArgs),
	NT =.. [F|NTArgs].

new_term(0,_TAs,NTArgs,NTArgs).
new_term(N0,TAs,Args,NTArgs) :-
	N0 > 0,
	member(T,TAs),
	N is N0 -1,
	new_term(N,TAs,[T|Args],NTArgs).
	
:- new_term(f/2,[f(a),a],NT).


%
% closure/2 <--
%

closure(State) :-
	eval_tas(State,TAs),
	eval_huniverse(State,HUs),
	eval_rules(State,Rs),
	\+ not_closure(TAs,HUs,Rs).

not_closure(TAs,HUs,Rs) :-
	member(F/A,HUs),
	new_term(F/A,TAs,NT),
	narrowTerm(NT,Rs,R),
	\+ member(R,TAs).

:- make_evalState([++'Q'(a)],[++equal(f(a),a)],State), closure(State).
:- \+ (make_evalState([++'Q'(a)],[++equal(f(a),a)],State), eval_tas(State,TAs), eval_huniverse(State,HUs), eval_rules(State,Rs), not_closure(TAs,HUs,Rs)).

%
% new_constant/2 <--
%

number_cons([],0).
number_cons([_F/A|HUs],N) :-
	\+ (A = 0),
	number_cons(HUs,N).
number_cons([_C/0|HUs],N) :-
	number_cons(HUs,N0),
	N is N0 + 1.

	
new_constant(HUs,D) :-
	number_cons(HUs,N0),
	N is N0 + 1,
	name(N,N_name),
	append("c","_",X),
	append(X,N_name,Y),
	name(D,Y).
	
:- new_constant([f/2,c/0],D).


%
% eval_theta/2 <--
%

eval_theta(State,TAs) :-
	closure(State),
	!,
	eval_tas(State,TAs).
eval_theta(State,Theta) :-
	eval_huniverse(State,HUs),
	eval_tas(State,TAs),
	new_constant(HUs,D),
	append([D],TAs,Theta).

:- make_evalState([++'Q'(a)],[++equal(f(a),a)],State), eval_theta(State,T).


%
% varlist
%

varlist([],[]).
varlist([L|Cs],Varlist) :-
	varlist_literal(L,Varlist0),
	varlist(Cs,Varlist1),
	merge(Varlist0,Varlist1,Varlist).

varlist_literal(Lit,Varlist) :-
	literal_atom(Lit,Atom),
	Atom =.. [_P|Args],
	varlist_args(Args,[],Varlist).

varlist_term(T,Varlist) :-
	T =.. [_F|TArgs],
	varlist_args(TArgs,[],Varlist).

varlist_args([],Varlist,Varlist).
varlist_args([V|Ts], Varlist0, Varlist) :-
	var(V),
	\+ memberNG_check(V,Varlist0),
	varlist_args(Ts,[V|Varlist0],Varlist).
varlist_args([V|Ts], Varlist0, Varlist) :-
	var(V),
	memberNG_check(V,Varlist0),
	varlist_args(Ts,Varlist0,Varlist).
varlist_args([T|Ts], Varlist0, Varlist) :-
	nonvar(T),
	varlist_term(T,Varlist1),
	merge(Varlist0,Varlist1,Varlist2),
	varlist_args(Ts,Varlist2,Varlist).
	
:- varlist_literal(++q(f(f(X,X),Y)),Var).
:- varlist([++'Q'(f(f(X))),--'P'(X,Y)],V), V=[X,Y].

%
% fill_varpos/4
%

fill_vars([],Clause,_Theta,Clause).
fill_vars([V|Varlist],Clause,Theta,GClause) :-
	member(T,Theta),
	V=T,
	fill_vars(Varlist,Clause,Theta,GClause).

:- fill_vars([X,Y],[++'Q'(f(f(X))),--'P'(X,Y)],[d,e],GC), GC=[++'Q'(f(f(e))),--'P'(e,d)].

%
% ground_projection
%

ground_projection_literal(Literal,Theta,GLiteral) :-
	varlist_literal(Literal,Varlist),
	fill_vars(Varlist,Literal,Theta,GLiteral).

ground_projection_clause(Clause,Theta,GClause) :-
	varlist(Clause,Varlist),
	fill_vars(Varlist,Clause,Theta,GClause).


:- ground_projection_literal(++q(f(f(X,X),Y)),[a,f(a),c_5],GC).
:- ground_projection_clause([++'Q'(f(f(X))),--'P'(X,Y)],[a,f(a),c_5],GC).

%--------------------------------------------------------------%
%                                                              %
%    general clause evaluation                                 %
%                                                              %
%--------------------------------------------------------------%

eval(FName,Clause):-
	atom(FName),
	!,
	eval2(FName,Clause).
eval(FName,_Clause) :-
	\+ atom(FName),
	write('Error: '),
	write(FName),
	write(' is not a proper filename'),
	nl.

eval2(FName,Clause) :-
	is_clause(Clause),
	!,
	eval3(FName,Clause).
eval2(FName,Literal) :-
	is_literal(Literal),
	!,
	eval3(FName,Literal).
eval2(_FName,Nonsense) :-
	write('Error: Sytax Error while parsing '),
	write(Nonsense),
	nl.

eval3(FName,Clause) :-
	read_arm_as_list(FName,X), 
	mb_parser(X,eval(Ats0,Eqs,HUs)),
	compSize(Eqs,Rules),
 	narrowing(Ats0,Rules,Ats),
	build_tas(Ats,TAs),
	evalState(eval(Ats,Rules,HUs,TAs),Clause),
	!,
	numbervars(Clause,0,_),
	write(Clause),
	write(' is true'),
	nl.	
eval(_FName,Clause):-
	numbervars(Clause,0,_),	
	write(Clause),
	write(' is false'),
	nl.

%-----------------------------------
%
% eval/3
%
%-----------------------------------

eval(Lits,Eqs,Clause) :-
	ground(Lits),
	is_neq_literals(Lits),
	is_geq_literals(Eqs),
	!,
	eval2(Lits,Eqs,Clause).
eval(Lits,Eqs,_Clause) :-
	(is_geq_literals(Eqs); \+ground(Lits)),
	!,
	write('Error: '),
	numbervars(Lits,0,_),
	write(Lits),
	write(' must be a set of non-equational literals').
eval(_Lits,Eqs,_Clause) :-
	write('Error: '),
	write(Eqs),
	write(' must be a set of equational literals'),
	nl.

eval2(Lits,Eqs,Clause) :-
	is_clause(Clause),
	!,
	eval3(Lits,Eqs,Clause).
eval2(Lits,Eqs,Literal) :-
	is_literal(Literal),
	!,
	eval3(Lits,Eqs,Literal).
eval2(_Lits,_Eqs,Nonsense) :-
	write('Error: Sytax Error while parsing '),
	numbervars(Nonsense,0,_),
	write(Nonsense),
	nl.

eval3(Lits,Eqs,Formula) :-
	make_evalState(Lits,Eqs,State),
	evalState(State,Formula),
	!,
	numbervars(Formula,0,_),
	write(Formula),
	write(' is true'),
	nl.	
eval3(_Lits,_Eqs,Formula) :-
	numbervars(Formula,0,_),
	write(Formula),
	write(' is false'),
	nl.	

evalState(State,Formula) :-
	ground(Formula),
	!,
        evalState_ground(State,Formula).
evalState(State,Formula) :-
	evalState_ng(State,Formula).

%
% 
% evalState_clause2 <--
%

evalState_ng(eval(Ats,Rs,HU0s,TAs),Literal) :-
	is_literal(Literal),
	extract_literal(Literal,HU1s),
	merge(HU0s,HU1s,HUs),
	literalNG_true(eval(Ats,Rs,HUs,TAs),Literal).
evalState_ng(eval(Ats,Rs,HU0s,TAs),Clause) :-
	is_clause(Clause),
	extract_fs(Clause,HU1s),
	merge(HU0s,HU1s,HUs),
	clauseNG_true(eval(Ats,Rs,HUs,TAs),Clause).

%
% literalNG_true/2
%

literalNG_true(State,Literal) :-
	\+ literalNG_false(State,Literal).

literalNG_false(State,Literal) :-
	eval_theta(State,Theta),
	ground_projection_literal(Literal,Theta,GLiteral),
	literalG_false(State,GLiteral).

%
% clauseNG_true/2
%

clauseNG_true(State,Clause) :-
	\+ clauseNG_false(State,Clause).

clauseNG_false(State,Clause) :-
	eval_theta(State,Theta),
	ground_projection_clause(Clause,Theta,GClause),
	clauseG_false(State,GClause).
	
% tests
%:-eval([++p(c_4,c_4,c_4)],[++equal(c,c_4),++equal(f(c_4),c_4),++equal(g(a),c_4),++equal(g(b),c_4),++equal(g(c_4),c_4)],++p(f(g(a)),f(g(b)),f(g(c)))).

%:-eval([++p(c_4,c_4,c_4)],[++equal(c,c_4),++equal(f(c_4),c_4),++equal(f(f(a)),c_4),++equal(f(f(b)),c_4),++equal(f(g(f(a))),c_4),++equal(f(g(f(b))),c_4),++equal(g(a),c_4),++equal(g(b),c_4),++equal(g(c_4),c_4),++equal(g(g(f(a))),c_4),++equal(g(g(f(b))),c_4)],++p(f(g(a)),f(g(b)),f(g(c)))).







