% ------------------------
% depthOrd(S,T) iff
%          depth(S) > depth(T) or
%          depth(S) = depth(T) and
%          depthLpo(S,T) 
% i.e. if both terms have equal depth than use a version of LPO that
% recursively applies depthOrd. 
% depthOrd needs a precedence on \Sigma
% 
% Remark: this ordering should fulfill the condition:
% depth(S) > depth(T) implies depthOrd(S,T) and
% is a CSO, TEST!
% -------------------------

:- module(depthOrd,[depth/2,
	depthOrd/2,
	depthOrdEq/2,
	include/2,
	sortFile/2]).
	
:- assertz(user:file_search_path(basic,'.')).

:- use_module(library(lists), [member/2]).
:- use_module(basic(order)).

%-------------------------
% depth(+S,-N).
% N is integer
% needs max/3
%-------------------------

depth(C,0) :-
    atom(C),
    !.
depth(V,0) :-
    var(V),
    !.
depth(T,N) :-
    T =.. [_F|Args],
    depthArgs(Args,N0),
    N is N0 +1.

depthArgs([], 0).
depthArgs([T|Ts], N) :-
    depth(T,N0),
    depthArgs(Ts,N1),
    N is max(N0,N1).

%-------------------------
% max(+A,+B,-Max).
% Max is integer.
% uses max/4 sorry the cuts are
% necessary. :-(
%-------------------------

max(A,B,Max) :-
    max(A,B,B,Max).

max(0,_B,Borig,Borig) :-
    !.
max(A,0,Borig,Max) :-
    !,
    Max is A + Borig.
max(A,B,Borig,Max) :-
    A1 is A - 1,
    B1 is B - 1,
    max(A1,B1,Borig,Max).

%--------------------------
% depthOrd(+S,+T).
% needs depth/2, 
%       depthOrd/4,
%       depthLpo/2,
%       ord_f/2
%--------------------------

depthOrd(S,T) :-
    depth(S,N),
    depth(T,M),
    depthOrd(S,T,N,M).

depthOrd(_S,_T,N,M) :-
    N > M.
depthOrd(S,T,N,M) :-
    N =:= M,
    depthLpo(S,T).

%-------------------------
% depthOrdEq(+S,+T).
% depthOrd + equal
%-------------------------

depthOrdEq(S,S).
depthOrdEq(S,T) :-
    depthOrd(S,T).

%--------------------------
% depthLpo(+S,+T).
% needs depthOrd/2,
%       all_Rs/2,
%       depthLex/2,
%       ord_f_trans/2,
%       equiv_f_trans/2
%
% depthLpo(S,T) :-
%     S =.. [_F|SArgs],
%     member(L,SArgs),
%     depthOrd(L,T).
%
% can be eliminated from the definition of depthOrd, because
% the above call of depthOrd/2 cannot be sucessfull, either 
% \tau(S) = \tau(T), then depthOrd(T,L) or
% \tau(S) > \tau(T), then depthLpo is never called.
%
% all_Rs(S,TArgs) in the second formula is allways true,
% hence redundant.
%
% depthLpo/2 needs not check any orderings
% between variables and non-variables, as those cannot
% be ordered by definition of depthOrd.
%--------------------------

%depthLpo(S,T) :-
%     S =.. [_F|SArgs],
%    member(L,SArgs),
%    depthOrd(L,T).

%depthLpo(S,T) :-
%    var(S),
%    var(T),
%    S==T.
depthLpo(S,T) :-
    nonvar(S),
    nonvar(T),
    S =.. [F|_SArgs],
    T =.. [G|_TArgs],
    ord_f_trans(F,G).
depthLpo(S,T) :-
    nonvar(S),
    nonvar(T),
    S =.. [F|SArgs],
    T =.. [F|TArgs],
    depthLex(SArgs,TArgs).


%-------------------------
% depthLex(+Ls,+Rs).
% needs depthOrd/2
%       quicksort(depth,_,_)/3 [sort.pl,order.pl]
%
% Remark:
% depthOrdEq is replaced by depthOrd as
% wrt. to the depth ord. an argument cannot be equal to
% the term T, otherwise the predicate depthLpo is not reached!
%-------------------------

depthLex(Ls, []) :-
    \+(Ls = []).
depthLex(Ls0, Rs0) :-
    quicksort(depth,Ls0,Ls),
    quicksort(depth,Rs0,Rs),
    depthLex2(Ls,Rs).

depthLex2([L|_Ls],[R|_Rs]) :-
    depthLpo(L,R).
depthLex2([L|Ls],[R|Rs]) :-
    L==R,
    depthLex(Ls,Rs).

%------------------------
% all_Rs(+S,+TArgs).
% needs depthOrd/2
%------------------------

all_Rs(_S,[]).
all_Rs(S,[R|Rs]) :-
    depthOrd(S,R),
    all_Rs(S,Rs).


%----------------------
% ord_f_trans(+F,+G).
% compute transitive closure of ord_f/2.
%----------------------

ord_f_trans(F,G) :-
    ord_f(F,G).
ord_f_trans(F,G):-
    ord_f(F,Between),
    ord_f_trans(Between,G).

%----------------------
% equiv_f_trans(+F,+G).
% compute reflexive, transitive closure of equiv_f/2.
%----------------------

equiv_f(F,F).

equiv_f_trans(F,G) :-
    equiv_f(F,G).
equiv_f_trans(F,G):-
    equiv_f(F,Between),
    equiv_f_trans(Between,G).

%-----------------------------
% include(+FILENAME,-Terms).
%-----------------------------

include(FName, Ts) :-
   current_input(OldStream),
   open(FName, read, Stream),
   read(Stream,T),
   include(Stream,T,Ts),
   close(Stream),
   set_input(OldStream),
   !.

include(_Stream, end_of_file, []) :-
   !.
include(Stream, T0, [T0|Ts]) :-
   read(Stream,T1),
   include(Stream, T1, Ts).


%------------------------------
% sortFile(+FILENAME, -Terms).
%------------------------------

sortFile(FName, Ts) :-
   include(FName, Ts0),
   quicksort(depthOrd,Ts0, Ts),
   printTs(Ts).

%------------------------------
% printTs(Ts).
%------------------------------

printTs([]) :-
   format("\n", []).
printTs([T|Ts]) :-
   format("~k < ", [T]),
   printTs(Ts).


%------------------------------
% Tests
%------------------------------

% ord_f declaration:

ord_f('P',f).
ord_f(f,g).
ord_f(g,h).
ord_f(h,b).
ord_f(b,a).

% test predicates
depthOrdTest(S,T) :-
    depthOrd(S,T),
    \+ depthOrd(T,S).

% hans' counterexample is no more valid:

:- depthOrd(g(a),b). 
:- depthOrd(f(g(a),h(a)),f(b,h(a))).

% extension to non-ground case

:- depthOrdTest('P'(X,f(X)),'P'(X,g(X))).
:- \+ (depthOrd(a,a)).
:- \+ (depthOrd(A,A)).
:- \+ (depthLpo(g(f(X),Y),g(Z,Y))).
:- \+ (depthLpo(f(X),Y)).
