% general.pl
%
% general predicates like append, select etc.
%
% contains:
%  	is_list/1,
%  	is_tail_variable_liste/1,
%     	is_list_of/2,
%  	is_non_negative_integer/1,
%       member/2 (library)
%       memberchk/2 (library)
%	memberNG_check/2
%	key_member/2,
%	key_memberNG_check/2.
%	append/3, (library)
%	select/3, (library)
%	selectNG/3,
%	delete_list/3,
%	delete_listNG/3,
%  	nth/3,
%  	reverse/2,
%     	list_to_varlist/2,
%     	length/3,
%     	length_difflist/3,
%       merge/3

:- module(general,[is_tail_variable_liste/1,
	is_list_of/2,
	is_non_negative_integer/1,
	memberNG_check/2,
	key_member/2,
	key_memberNG_check/2,
	selectNG/3,
	delete_list/3,
	delete_listNG/3,
	list_to_varlist/2,
	my_length/3,
     	length_difflist/3,
 	merge/3]).

%      	

:- use_module(library(lists)).

% difference list operator
:-op(640,xfx,\).

% is_tail_variable_liste(?Xs) <-- 
% 			Xs is a list with tail variable

is_tail_variable_liste(Xs) :-
	var(Xs).
is_tail_variable_liste(Xs) :-
	nonvar(Xs),
	Xs = [_|_].

% is_list_of(+Xs, +PredName) <--
%        Xs is a list of PredName's

is_list_of(Xs, _PredName) :-
	nonvar(Xs),
	Xs = [].
is_list_of(Xs0, PredName) :-
	nonvar(Xs0),
	Xs0 = [X|Xs],
	Goal =.. [PredName,X],
	Goal,
	is_list_of(Xs,PredName).

% is_non_negative_integer(+N) <--
% 			N element of N not N0

is_non_negative_integer(N) :-
	integer(N),
	N > 0.

% memberNG_chk(+X, +Xs) <--
% 			same as member_check, but logical variables are treated as objects

memberNG_check(X, [Y|_Xs]) :-
	X==Y.
memberNG_check(X, [Y|Xs]) :-	
	\+ X==Y,
	memberNG_check(X,Xs).

% key_member(?Key, +Xs) <--
% 			Key is a member of the pairlist Xs

key_member(Key, [Key-_Value|_Xs]).
key_member(Key, [_KeyValue|Xs]) :-
	key_member(Key,Xs).

% key_memberNG_check(+Key, +Xs) <--
% 			same as key-member, but treats logical variables as objects

key_memberNG_check(Key, [Key1-_Value|_Xs]) :-
	Key==Key1.
key_memberNG_check(Key, [Key1-_Value|Xs]) :-
	\+ Key==Key1,
	key_memberNG_check(Key,Xs).


% selectNG(+X, +Xs, -Ys) <--
% 			same as select, but logical variables are treated as objects

selectNG(X, [X1|Xs], Xs) :-
	X==X1.
selectNG(X, [Y|Xs], [X|Ys]) :-
	\+ X==Y,
	selectNG(X, Xs,Ys).

% delete_list(+Xs, +Ys, -Zs) <-- 
% 			Zs is Ys without Xs

delete_list(Xs,[],Xs).
delete_list([X|Xs],Ys0,Zs) :-
	Ys0 = [_|_],
	select(Ys0,X,Ys),
	delete_list(Xs,Ys,Zs).
delete_list([X|Xs],Ys,[X|Zs]) :-
	Ys = [_|_],
	\+ member(X,Ys),
	delete_list(Xs,Ys,Zs).

% delete_listNG(+Xs, +Ys, -Zs) <--
% 			same as delete_list, but logical variables are treated as objects

delete_listNG(Xs,[],Xs).
delete_listNG([X|Xs],Ys0,Zs) :-
	Ys0 = [_|_],
	select(Ys0,Y,Ys),
	X==Y,
	delete_listNG(Xs,Ys,Zs).
delete_listNG([X|Xs],Ys,[X|Zs]) :-
	Ys = [_|_],
	\+ memberNG_check(X,Ys),
	delete_listNG(Xs,Ys,Zs).

% list_to_varlist(+Xs, -Ys) <--
%       deletes all non variables form Xs 

list_to_varlist([], []).
list_to_varlist([X|Xs], [X|Ys]) :-
	var(X),
	list_to_varlist(Xs,Ys).
list_to_varlist([X|Xs], Ys) :-
	nonvar(X),
	list_to_varlist(Xs,Ys).

% my_length(+Xs, +PredName, -Length) <--
%      length is the length of a list of PredName-structs
%      for these structs their has to be an rule PredName(length,Struct,StructLength)

my_length(Xs, PredName, Length) :-
	prolog:current_predicate(PredName,_Goal),
	my_length_1(Xs,PredName,Length).

my_length_1([], _PredName, 0).
my_length_1([X|Xs], PredName, Length) :-
	Goal =.. [PredName,X,StructLength],
	Goal,
	my_length_1(Xs,PredName,Length0),
	Length is Length0 + StructLength.

% length_difflist(+Xs1, +Xs2, Length) <--
%      length of a difference list


length_difflist(Xs1, Xs2, 0) :-
	nonvar(Xs1),
	nonvar(Xs2),
	Xs1 = Xs2.
length_difflist(Xs1, Xs2, 0) :-
	var(Xs1),
	var(Xs2),
	Xs1 == Xs2.
length_difflist(Xs0, Xs2, Length) :-
	nonvar(Xs0),
	Xs0 = [_X|Xs1],
	length_difflist(Xs1,Xs2,Length0),
	Length is Length0 + 1.

% merge(+Xs,+Ys,-Zs) <--
%             merge Xs, Ys to Zs

merge(Xs,Ys,Zs) :-
	append(Xs,Ys,Zs0),
	remove_duplicates(Zs0,Zs).

:- merge([a,b,c],[a,b],X).





