% datastr.pl
% 
% predicates to implement the used datastructures for equations, rules, already
% found critical pairs and already checked terms
% 

:- module(mi_datastr,[is_2_plus_2_tree/1,
	is_value/1,
	getValue/3,
	addClosedPair/3,
	mapListTree/2,
	is_equations/1,
	is_rules/1,
	is_visited_cps/1,
	is_equation_queue/1,
	is_rule_queue/1,
	is_old_queue/1,
	is_trs_state/1,
	is_dbase/1,
	mapDatabase/5]).

:- assertz(user:file_search_path(basic,'.')).
:- assertz(user:file_search_path(trs,'.')).

:- use_module(basic(general)).
:- use_module(basic(term)).
:- use_module(trs(mi_rep)).

% defines a tree to represent to visited critical pairs

is_2_plus_2_tree(Xs) :-
	var(Xs).
is_2_plus_2_tree(node(Key,Value,L,R)) :-
	integer(Key),
	is_value(Value),
	is_2_plus_2_tree(L),
	is_2_plus_2_tree(R).

is_value(Value) :-
	is_tail_variable_liste(Value).

% getValue(+Key, +Tree, -Value) <--
% 	get the to Key coresponding value 

getValue(Key, node(Key,Value,_,_), Value).
getValue(Key, node(Higher,_,L,_), Value) :-
	integer(Higher),
	Higher > Key,
	getValue(Key,L,Value).
getValue(Key, node(Lower,_,_,R), Value) :-
	integer(Lower),
	Key > Lower,
	getValue(Key,R,Value).	

% addClosedPair(+Key, +RuleNr, ?Tree) <--
% 	Key and RuleNr form a pair of an already found critical pair 
% 	NewTree is Tree with the additional information that Key and RuleNr have been
% 	visited

addClosedPair(Key, RuleNr, node(Key,Value,_,_)) :-
	insertDs(RuleNr,Value).
addClosedPair(Key, RuleNr, node(Higher,_,L,_)) :-
	integer(Higher),
	Higher > Key,
	addClosedPair(Key,RuleNr,L).
addClosedPair(Key, RuleNr, node(Lower,_,_,R)) :-
	integer(Lower),
	Key > Lower,
	addClosedPair(Key,RuleNr,R).

% mapListTree(+Keys, -Tree) <--
% 	represent a list of keys as a 2+2 tree

mapListTree([Key|Keys],Tree) :-
	mapListTree(Key,Keys,Tree).

mapListTree(Key,[],node(Key,_Value,_L,_R)).
mapListTree(Key, [NextKey|Keys], node(Key,_Value,_L,R)) :-
	NextKey > Key,
	mapListTree(NextKey,Keys,R).
mapListTree(Key, [NextKey|Keys], node(Key,_Value,L,_)) :-
	Key >= NextKey,
	mapListTree(NextKey,Keys,L).

:- is_2_plus_2_tree(node(2,[1,2,3],node(1,_Value,_L1,_R1),_R2)).
:- mapListTree([2,4,3],Tree).
:- mapListTree([2,4,3],Tree),addClosedPair(3,4,Tree).
:- getValue(1,node(2,[1,2|_],node(1,_,_,_),node(3,[3|_],_,_)),Value).

% is_equations(Equations) <--
%	Equations is a list of equations

is_equations([]).
is_equations([E|Equations]) :-
	is_equation(E),
	is_equations(Equations).

% is_rules(Rules) <--
%	Rules is a list of rules

is_rules([]).
is_rules([R|Rules]) :-
	is_rule(R),
	is_rules(Rules).

% is_visited_cps(Cps) <--
%	Cps is a list of visited critical pairs

is_visited_cps([]).
is_visited_cps([Cp|Cps]) :-
	is_visited_cp(Cp),
	is_visited_cps(Cps).

% is_equation_queue(EquationQueue) <--
% 	EquationQueue is a queue of equations

is_equation_queue(queue(0,Xs,Xs)).
is_equation_queue(queue(N,[E-Index|Xs],Ys)) :-
	integer(N),
	integer(Index),
	N > 0,
	is_equation(E),
	N0 is N - 1,
	is_equation_queue(queue(N0,Xs,Ys)).

% is_rule_queue(RuleQueue) <--
% 	RuleQueue is a queue of rules

is_rule_queue(queue(0,Xs,Xs)).
is_rule_queue(queue(N,[R-Index|Xs],Ys)) :-
	integer(N),
	integer(Index),
	N > 0,
	N0 is N - 1,
	is_rule(R),
	is_rule_queue(queue(N0,Xs,Ys)).

% is_old_queue(OldQueue) <--
% 	OldQueue is a queue of already visited - old - critical pairs

is_old_queue(queue(0,Xs,Xs)).
is_old_queue(queue(N,[Cp-Index|Xs],Ys)) :-
	integer(N),
	integer(Index),
	N > 0,
	N0 is N - 1,
	is_critical_pair(Cp),
	is_old_queue(queue(N0,Xs,Ys)).

% is_trs_state(Trs_State) <--
%	defines the states of the completion procedure

is_trs_state(trs(RedOrd,EquationQueue,RuleQueue0,OldQueue,ClosedTree)) :-
	is_order(RedOrd),
	is_equation_queue(EquationQueue),
	is_rule_queue(RuleQueue0),
	is_old_queue(OldQueue),
	is_2_plus_2_tree(ClosedTree).

% is_dbase(Database) <--
%	specifies the datastructure to store the equations / rules and
%	already visited critical pairs
%
is_dbase(dbase(EquationQueue,RuleQueue,OldQueue,ClosedTree)) :-
	is_equation_queue(EquationQueue),
	is_rule_queue(RuleQueue),
	is_old_queue(OldQueue),
	is_2_plus_2_tree(ClosedTree).

% mapDatabase(DataBase,EquationQueue,RuleQueue,OldQueue,ClosedTree) <--
%		   form a database as specified by is_dbase/1

mapDatabase(dbase(EQueue,RQueue,OQueue,CTree),EQueue,RQueue,OQueue,CTree) :-
	is_equation_queue(EQueue),
	is_rule_queue(RQueue),
	is_old_queue(OQueue),
	is_2_plus_2_tree(CTree).

:- mapDatabase(dbase(queue(3,[(1*x=x)-1,(x*1=x)-2,(i(x)*(x*y)=y)-3|Xs],Xs),queue(0,Ys,Ys),queue(0,Zs,Zs),node(1,_Value,_,_)),E,R,O,C),E=queue(3,[(1*x=x)-1,(x*1=x)-2,(i(x)*(x*y)=y)-3|Xs],Xs),R=queue(0,Ys,Ys),O=queue(0,Zs,Zs),C=node(1,_Value,_,_).
