% trs_iDJ.pl
%
% implements a metainterpreter to handle the reduction steps
% in [DJ90]
% in this interpreter the reduction steps have the (almost) the same
% syntax as in [DJ90], so one can at first manipulate the complition procedure
% by manipulating of these reductions steps. E.g. one might change the order, 
% merge two steps or define additional conditions.
% In addition to these steps it is possible to define the predicate cps/2 by 
% a specialized language, given in is_cps_body(X).
% If there is no user-defined predicate cps/2 the critical pairs are searched
% between the smallest rules

:- module(mi_idj,[is_step/1,
	iDJ/3,
	iDJ_next/5,
	is_condition/1,
	condition/5,
	is_cps_body/1,
	critical_pair/3,
	mapQueuesStep/3,
	mapStepQueues/3]).

:- assertz(user:file_search_path(basic,'.')).
:- assertz(user:file_search_path(trs,'.')).

:- use_module(basic(queue)).
:- use_module(basic(order)).
:- use_module(trs(mi_datastr)).
:- use_module(trs(mi_pdb)).
:- use_module(trs(mi_rord)).
:- use_module(trs(mi_unif)).
:- use_module(trs(mi_cps)).

% iDJ(+State0,+RedOrder,State) <--
%	 State0 is the state before the reduction step
%	 RedOrder is the used reduction ordering
%	 State is the state after the reduction step

% defintion of a reduction step - the syntax of LangDJ

is_step((EQueue,RQueue)) :-
	is_step_equation_queue(EQueue),
	is_step_rule_queue(RQueue).

is_step_equation_queue(queue(0,Xs,Xs)).
is_step_equation_queue(EQueue) :-
	EQueue = queue(N,_,_),
	N > 0,
	is_equation_queue(EQueue).
is_step_equation_queue(E + EQueue) :-
	is_equation(E),
	is_equation_queue(EQueue).

is_step_rule_queue(queue(0,Xs,Xs)).
is_step_rule_queue(RQueue) :-
	RQueue = queue(N,_,_),
	N > 0,
	is_rule_queue(RQueue).
is_step_rule_queue(R + RQueue) :-
	is_rule(R),
	is_rule_queue(RQueue).

% iDJ(Trs_State0,Trs_State) <--
% 			Trs_State is a trs_state as defined in mi_rep.pl gained by
% 			reduction step from Trs_State0

iDJ(Database0,RedOrder,Database) :-
	mapDatabase(Database0,EQueue0,RQueue0,OQueue0,ClosedTree),
	iDJ_next(RedOrder,EQueue0==>EQueue,RQueue0==>RQueue,OQueue0==>OQueue,ClosedTree),
	mapDatabase(Database,EQueue,RQueue,OQueue,ClosedTree).

iDJ_next(RedOrd,EQueue0==>EQueue,RQueue0==>RQueue,OldCps,ClosedTree) :-
	Step0 & Condition ==> Step,
	mapQueuesStep(EQueue0,RQueue0,Step0),
	projectionStepRQueue(Step0,RQueue1),
	condition(Condition,RedOrd,RQueue1,OldCps,ClosedTree),
	mapStepQueues(Step,EQueue,RQueue).

% all possible conditions

is_condition((_S > _T)).
is_condition((_S >= _T)).
is_condition((_Rule1 >- _Rule2)).
is_condition(cps(_S,_T)).
is_condition(rewrite(_S,_T)).
is_condition(rewrite(_S,_T,_RewRule)).
is_condition(equivalent(_S,_T)).

% conditon(Condition,RedOrder,RQueue,OldQueue0==>OldQueue,ClosedTree) <--
%	 		Condition is the if-part of the reduction steps
%	 		RedOrder is used in the orient-step

condition((S > T),RedOrder,_RQueue,OldQueue==>OldQueue,_ClosedTree) :-
	order(RedOrder,S,T).
condition((S >= T),RedOrder,_RQueue,OldQueue==>OldQueue,_ClosedTree) :-
	order(RedOrder,S,T).
condition((Rule1 >- Rule2),RedOrder,_RQueue,OldQueue==>OldQueue,_ClosedTree) :-
	ruleOrd(RedOrder,Rule1,Rule2).
condition(cps(Cp_1,Cp_2),_RedOrder,RQueue,OldQueue0==>OldQueue,ClosedTree) :-
	cps(Cp_1,Cp_2,TRNr,RRNr) <= Body,
	critical_pair(Body,RQueue,ClosedTree),
	\+ redundantCp(Cp_1,Cp_2,OldQueue0),
	\+ redundantRule(Cp_1,Cp_2,RQueue),
        queueLast_Index(OldQueue0,cp(Cp_1,Cp_2),(TRNr:RRNr),OldQueue).
condition(rewrite(S,T),_RedOrder,RQueue,OldQueue==>OldQueue,_ClosedTree) :-
	queueMember(RewRule-_RRuleNr,RQueue),
	rewrite(S,RewRule,_Pos,T).
condition(rewrite(S,T,RewRule),_RedOrder,RQueue,OldQueue==>OldQueue,_ClosedTree) :-
	queueMember(RewRule-_RRuleNr,RQueue),
	rewrite(S,RewRule,_Pos,T).
condition(equivalent(S,T),_RedOrder,_RQueue,OldQueue==>OldQueue,_ClosedTree) :-
        ground_to_nonground(S,S_ng,[],Di),
        ground_to_nonground(T,T_ng,Di,_Di1),
        unify(S_ng,T_ng).
condition(A & B,RedOrder,RQueue,OldQueue0==>OldQueue,ClosedTree) :-
	condition(A,RedOrder,RQueue,OldQueue0==>OldQueue1,ClosedTree),
	condition(B,RedOrder,RQueue,OldQueue1==>OldQueue,ClosedTree).

% critical_pair(Body) <--
%	 Body is the body of the user-defined cps/2 to obtain a nice
%	 possibilty to defined the search for critical pairs 

is_cps_body(smallPair(_,_)).
is_cps_body(cp(_,_,_,_)).

critical_pair(smallPair(TopRule,RewRule,TRNr,RRNr), RQueue, ClosedTree) :-
	smallestRulePair(RQueue,ClosedTree,TopRule,RewRule,TRNr,RRNr).
critical_pair(cp(TopRule,RewRule,Cp_1,Cp_2), _RQueue, _ClosedTree) :-
	cp(TopRule,RewRule,cp(Cp_1,Cp_2)).
critical_pair(A & B,RQueue,ClosedTree) :-
	critical_pair(A,RQueue,ClosedTree),
	critical_pair(B,RQueue,ClosedTree).

% mapQueuesStep(+EQueue, +RQueue, -Step) <--
% 			makes a Step-structure from EQueue and RQueue

mapQueuesStep(EQueue,RQueue,(EQueue;RQueue)).
mapQueuesStep(EQueue1,RQueue,((Equation + EQueue);RQueue)) :-
	queueSelect(EQueue1,Equation-_Index,EQueue). 
mapQueuesStep(EQueue,RQueue1,(EQueue;(Rule + RQueue))) :-
	queueSelect(RQueue1,Rule-_Index,RQueue). 

:- mapQueuesStep(queue(1,[(a*1=a)-1|Xs],Xs),queue(1,[(a*1=>a)-1|Ys],Ys),(E+Es;R)).

% mapStepQueues(+Step, -EQueue, -RQueue) <--
% 			makes EQueue and RQueue from a Step-structure 

mapStepQueues((EQueue;RQueue), EQueue, RQueue) :-
	decompose(no,EQueue,RQueue).
mapStepQueues((EquationUnion;RQueue), EQueue, RQueue) :-
	decompose(new(Equation + EQueue1),EquationUnion,RQueue),
	queueLast_number(EQueue1,Equation,EQueue,_). 
mapStepQueues((EquationUnion;RQueue), EQueue, RQueue) :-
	decompose(old(EQueue),EquationUnion,RQueue).
mapStepQueues((EQueue;RuleUnion), EQueue, RQueue) :-
	decompose(new(Rule + RQueue1),RuleUnion,EQueue),
	queueLast_number(RQueue1,Rule,RQueue,_). 
mapStepQueues((EQueue;RuleUnion), EQueue, RQueue) :-
	decompose(old(RQueue),RuleUnion,EQueue).

decompose(no,XQueue,YQueue) :-
	XQueue = queue(_,_,_),
	YQueue = queue(_,_,_).
decompose(new(X + XQueue),X + XQueue,YQueue) :-
	XQueue = queue(_,_,_),
	YQueue = queue(_,_,_),
	\+ redundant(X,XQueue),
	\+ redundant(X,YQueue).
decompose(old(XQueue),X + XQueue,YQueue) :-
	XQueue = queue(_,_,_),
	YQueue = queue(_,_,_),
	redundant(X,XQueue).
decompose(old(XQueue),X + XQueue,YQueue) :-
	XQueue = queue(_,_,_),
	YQueue = queue(_,_,_),
	redundant(X,YQueue).

:- decompose(X,(1*x=a)+queue(0,Xs,Xs),queue(1,[1*y=>a|Ys],Ys)).
:- mapStepQueues(((1*u=u)-1+queue(0,Xs,Xs);queue(3,[(1*u=>u)-1,(i(x)*(x*y)=>y)-2,(i(1)=>1|Ys)-3],Ys)),E,R).

% projectionStepRQueue(+Step,-RQueue) :-
% 			RQueue is rules part of the step-structure Step

projectionStepRQueue((_;RQueue), RQueue) :-
	decompose(no,RQueue).
projectionStepRQueue((_;RuleUnion), RQueue) :-
	decompose(yes(_Rule + RQueue),RuleUnion).

decompose(no,XQueue) :-
	XQueue = queue(_,_,_),
	is_queue(XQueue).
decompose(yes(X + XQueue),X + XQueue) :-
	is_queue(XQueue).

:- projectionStepRQueue((queue(0,Xs,Xs);(c=>d)+queue(1,[a=>b|Xs],Xs)),queue(1,[a=>b|Xs],Xs)).

% reduction steps defined in LangDj

( (S=T)+E ; R )   & equivalent(S,T) ==> ( E ; R ).
( (S=T)+E ; R )   & rewrite(S,U)    ==> ( (U=T)+E ; R ).
( (S=T)+E ; R )   & rewrite(T,U)    ==> ( (S=U)+E ; R ).
( (S=T)+E ; R )   & (S >= T)        ==> ( E ; (S=>T)+R ).
( (S=T)+E ; R )   & (T > S)         ==> ( E ; (T=>S)+R ).
( E ; (S=>T)+R )  & rewrite(T,U)    ==> ( E ; (S=>U)+R ).
( E ; (S=>T)+R )  & rewrite(S,U,RewRule) & (S=>T) >- RewRule  ==> ( (U=T)+E ; R ).
( E ; R )         & cps(S,T)        ==> ( (S=T)+E ; R ).

cps(S,T,R1Nr,R2Nr) <= smallPair(Rule1,Rule2,R1Nr,R2Nr) & cp(Rule1,Rule2,S,T).
%tests

%condition
:- mapListTree([1,2],CTree), condition(cps(Cp_1,Cp_2),_,queue(2,[(1*x=>x)-1,(i(x)*(x*y)=>y)-2|Xs],Xs),queue(0,Zs,Zs)==>OQueue,CTree).
:- condition(rewrite(i(1)*u,U,RewRule) & ((i(1)*u=>u) >- RewRule),sizeOrd,queue(1,[(i(1)=>1)-1|Xs],Xs),_,_).
:- condition(rewrite(i(1)*u,U,RewRule),sizeOrd,queue(1,[(i(1)=>1)-1|Xs],Xs),_,_),condition(((i(1)*u=>u) >- RewRule),sizeOrd,queue(1,[(i(1)=>1)-1|Xs],Xs),_,_).

%idj
:- iDJ_next(sizeOrd,queue(0,Xs,Xs)==>EQueue,queue(2,[(1*x=>x)-1,(i(x)*(x*y)=>y)-2|Ys],Ys)==>RQueue,queue(0,Zs,Zs)==>OQueue,node(1,_Value1,_,node(2,_Value2,_,_))).

%general
:- Step0 & Condition ==> Step,mapQueuesStep(queue(4,[(1=1)-1,(1*a = a)-2,(a*1 = a)-3,(i(a)*(a*b)=b)-4|Xs],Xs),queue(4,[(1*x=>x)-1,(i(x)*(x*y)=>y)-2,(i(1)*u=>u)-3,(i(1)=>1)-4|Ys],Ys),Step0),mapStepQueues(Step,EQueue,RQueue).
:- Step0 & Condition ==> Step,mapQueuesStep(queue(0,Xs,Xs),queue(1,[1*x=>x|Ys],Ys),Step0),mapStepQueues(Step,EQueue,RQueue).


