
% mi_size.pl
%
% size orderings 

:- module(mi_size,[sizeOrd/2,
	strictSizeOrd/2,
	size/2,
	sizeOrdWeight/3,
	strictSizeOrdWeight/3,
	sizeWeight/3]).

:- assertz(user:file_search_path(basic,'.')).
:- assertz(user:file_search_path(trs,'.')).

:- use_module(library(lists),[member/2]).
:- use_module(trs(mi_var)).
:- use_module(trs(mi_ford)).

% Definition: size ordering ( >=size )
%
% size(S) is the number of funtion symbols in S
% 
% S >=size T iff 
%	size(S) >= size(T);
% 

% sizeOrd(+S,+T) <-- S >=size T

sizeOrd(S,T) :-
	size(S,S1),
	size(T,T1),
	S1 >= T1.

strictSizeOrd(S,T) :-
	size(S,S1),
	size(T,T1),
	S1 > T1.

% size(+Term,?Size, -Ops) <-- 
%	Size is the number of function symbols in term
%	Ops is a difference structure storing the already
%	found function symbols

size(Term,Size) :-
	size(Term,Size,[]-_Ops).

size(V,0,Ops-Ops) :-		        % variables have weight 0
	is_user_var(V),
	!.
size(C,0,Ops-Ops) :-		        % constants have weight 0
	non_user_var(C),
	atomic(C),
	!.
size(V,0,Ops-Ops) :-
        var(V),
        !.
size(Term,Size,Ops0-Ops) :-	  % decompose term and handle the 
	Term =.. [F|TArgs],          % found function symbol
	\+ member(F,Ops0),
	!,
	size_list(TArgs,Size0,[F|Ops0]-Ops),
	Size is Size0 + 1.
size(Term,Size,Ops0-Ops) :-	  % the function symbol was already found
	Term =.. [_|TArgs],
	size_list(TArgs,Size,Ops0-Ops).

size_list([],0,Ops-Ops).
size_list([X|Xs],Size,Ops0-Ops) :-
	size(X,Size0,Ops0-Ops1),
	size_list(Xs,Size1,Ops1-Ops),
	Size is Size0 + Size1.

:- size(f(a),1).
:- size(f(x,f(a,y)),S).
:- size(1*a,S,Ops).

% Definition: size ordering with special weights ( >=sizeW )
%
% sizeW(S) is the sum of the weights of funtion symbols in S plus the
% number of variables in S times the lowest weight of an function symbol

% S >=sizeW T iff
%	sizeW(S) >= sizeW(T);

% sizeOrdWeight(+Ws,+S,+T) <-- 
%	S >=sizeW T 
%	Ws is the list of function symbol weights

sizeOrdWeight(Ws,S,T) :-
	sizeWeight(Ws,S,Size_S),
	sizeWeight(Ws,T,Size_T),
	Size_S >= Size_T.

% Definition: strict size ordering with special weights ( >sizeW )
%
% >=lex is the induced lexicographic ordering
% sizeW(S) is the sum of the weights of funtion symbols in S plus the
% number of variables in S times the lowest weight of an function symbol

% S >sizeW T iff
% 
%  (i)  sizeW(S) > sizeW(T);
%  (ii) sizeW(S) = sizeW(T) and
%	   the arguments of S are greater than the arguments of T in the
%	   induceded lexicographic ordering

% strictSizeOrdWeight(+Ws,+S,+T) <-- 
%		 S >=sizeW T 
%		 Ws is the list of function symbol weights

strictSizeOrdWeight(Ws,S,T) :-
	sizeWeight(Ws,S,Size_S),
	sizeWeight(Ws,T,Size_T),
	Size_S > Size_T.
strictSizeOrdWeight(Ws,S,T) :-
	sizeWeight(Ws,S,Size_S),
	sizeWeight(Ws,S,Size_T),
	Size_S =:= Size_T,
	compound(S), 
	compound(T),
	S =.. [F|Arg_Ss],
	T =.. [F|Arg_Ts],
	strictLexOrd(order(sizeOrdWeight,Ws),Arg_Ss,Arg_Ts).

% sizeWeight(Ws,Term,Size,Ops) <-- 
%	Size is the number of different function symbols in term
%	Ws a list of weights of function symbols
%	Ops is a difference structure storing the already
%	found function symbols

sizeWeight(Ws,Term,Size) :-
	sizeWeight(Ws,Term,Size,[]-_Ops).

sizeWeight(Ws,V,Min,Ops-Ops) :-	      % variables are weighted
	is_user_var(V),			            % with the smallest weight of
	!,	                                 % a function symbol
	minimalWeight(Ws, Min).	       
sizeWeight(_Ws,C,0,Ops-Ops) :-	      % weight of constants is 0
	non_user_var(C),
	atomic(C),
	!.
sizeWeight(Ws,Term,Size0,Ops0-Ops) :-  % decompose the term and manipulate
	Term =.. [F|TArgs],                 % the found function symbol
	\+ member(F,Ops0),
	!,
	sizeWeight_list(Ws,TArgs,Size1,[F|Ops0]-Ops),
	weight(F,Ws,Weight),
	Size0 is Size1 + Weight.
sizeWeight(Ws,Term,Size,Ops0-Ops) :-   % the function symbol was already found
	Term =.. [_|TArgs],
	sizeWeight_list(Ws,TArgs,Size,Ops0-Ops).

sizeWeight_list(_Ws,[],0,Ops-Ops).
sizeWeight_list(Ws,[T|TArgs],Size,Ops0-Ops) :-
	sizeWeight(Ws,T,Size0,Ops0-Ops1),
	sizeWeight_list(Ws,TArgs,Size1,Ops1-Ops),
	Size is Size0 + Size1.

% minimalWeight(Ws, Min) <-- 
%	Min is the minimal weight of a function symbol
%	specified in Ws

minimalWeight([(_F = W)|Ws], Min) :- 
	minimalWeight(W,Ws,Min). 

minimalWeight(Min,[],Min).
minimalWeight(OldMin,[(_F = W)|Ws],Min) :-
	OldMin =< W, 
	!,
	minimalWeight(OldMin,Ws,Min).
minimalWeight(_OldMin,[(_F = W)|Ws],Min) :-   
	minimalWeight(W,Ws,Min).

:- minimalWeight([(i = 2),(* = 1)],1).
:- sizeWeight([(f=2)],f(x),N).
:- sizeWeight([(f=2)],f(x,f(y,z)),N), N=8.
:- strictSizeOrdWeight([(i = 2),(* = 1)],x*i(x),1).
:- sizeOrdWeight([(i = 2),(* = 1)],(x*y)*z,x*(y*z)).
