
% mi_unif.pl
%
% syntactic matching and unification predicates

:- module(mi_unif,[is_unifier/1,
	unifier/3,
	match/2,
	unify/2,
	ground_to_nonground/2,
	ground_to_nonground/4,
	nonground_to_ground/2,
	nonground_to_ground/4]).

:- assertz(user:file_search_path(basic,'.')).
:- assertz(user:file_search_path(trs,'.')).

:- use_module(basic(term)).
:- use_module(trs(mi_var)).

% operator definition
:- op(650,xfx,<-).

is_unifier([]).
is_unifier([X <- Var|Unifier]) :-
	var(X),
	var(Var),
	is_unifier(Unifier).
is_unifier([X <- UnifyTerm|Unifier]) :-
	var(X),
	functor(UnifyTerm,_,_),
	is_unifier(Unifier).

% unifier(+X, +Unifier, -Substitution) <--
%        returns the term to substitute for X in Unifier

unifier(X, [Y <- UnifyTerm|_Unifier], UnifyTerm) :-
	Y==X.
unifier(X, [Y <- _UT|Unifier], UnifyTerm) :-
	\+ Y==X,
	unifier(X,Unifier,UnifyTerm).

% represent unifiers as difference lists

unifier(X, [Y <- UnifyTerm|_Unifier1], _Unifier2, UnifyTerm) :-
	Y==X.
unifier(X, [Y <- _UT|Unifier1], Unifier2, UnifyTerm) :-
	\+ Y==X,
	unifier(X,Unifier1,Unifier2,UnifyTerm).

% Definition:
% S matches T; S = sigma(T), sigma is an idempotent substitution, of
% which binds variables in S

match(S,T) :-
	match(S,S,T).

% match(+Term, +S, +T) <-- S = sigma(T)
%	Term is used to control the match

match(Original, S, T) :-	   
	var(S),
	match_var(Original,T,S).
match(Original, S, T) :-
	nonvar(S), 
	match_term(Original,T,S).

match_var(_Original, X, Var) :-
	Var==X.
match_var(Original, X, Var) :-
	var(X),
	\+ occurs_in(X,Original),
	X=Var.
match_var(Original,X,Var) :-
	nonvar(X),
	\+ occurs_in(Var,Original),
	Var=X.

match_term(Original,X,Term) :-
	var(X),
	match_var(Original,Term,X).
match_term(Original,X,Term) :-
	nonvar(X),
	X =.. [F|TArgs],
	Term =.. [F|SArgs],
	match_list(Original,SArgs,TArgs).

match_list(_Original, [], []).
match_list(Original, [S|Ss], [T|Ts]) :-
	match(Original,S,T),
	match_list(Original,Ss,Ts).

:- \+ match(f(A,B,C),f(X,Y,A)).
:- \+ match(f(X,Y),f(Z,Z)).
:- match(X*1,Y*1).

% unify(+S, +T) <--
%	S unifies syntactically with T
%	an occure check is made

unify(S, T) :-
	var(S), 
	unify_var(T,S).
unify(S, T) :-
	nonvar(S), 
	unify_term(T,S).

unify_var(X, Var) :-
	var(X),
	Var=X.
unify_var(X, Var) :-
	nonvar(X),
	\+ occurs_in(Var,X),
	Var=X.
unify_term(X, Term) :-
	var(X),
	unify_var(Term,X).
unify_term(X, Term) :-
	nonvar(X),
	X =.. [F|XArgs],
	Term =.. [F|TArgs],
	unify_list(XArgs,TArgs).

unify_list([], []).
unify_list([S|SArgs], [T|TArgs]) :-
	unify(S,T),
	unify_list(SArgs,TArgs).

:- unify(X*Y,1*U).
:- \+ unify(f(b,a),f(X,X)).
:- \+ unify(f(X,Y,Z),f(X,Y)).
:- unify(f(a,X),f(X,a)).

% list_unify(+Xs) <--
% 			unify the elements of Xs

list_unify([]).
list_unify([X|Xs]) :-
	list_unify_1(Xs,X).

list_unify_1([], _X).
list_unify_1([Y|Xs], X) :-
	unify(X,Y),
	list_unify_1(Xs,X).

:- list_unify([Y,f(X),f(f(a))]).

% ground_to_nonground(+Ground, -NonGround) <-- 
% 	maps a term with user-variables to a term with logical variables
% 	to simplify unification with this term

ground_to_nonground(Ground, NonGround) :-
	ground_to_nonground(Ground,NonGround,[],_VarDicts).

% ground_to_nonground(+Ground, -NonGround, +VarDicts0, -VarDicts) <--
%	VarDicts stores the variable assignments

ground_to_nonground(Ground, Var, VarDicts0, VarDicts) :-   % user-variables map to
	is_user_var(Ground),                                    % logical variables
	var(Var),                                               % generate a logical variable
	lookupGenerate(Ground,Var,VarDicts0,VarDicts).
ground_to_nonground(Var1, Var2, VarDicts0, VarDicts) :-
        var(Var1),
        var(Var2),
        lookupGenerate(Var1,Var2,VarDicts0,VarDicts).
ground_to_nonground(Ground, NonGround, VarDicts0, VarDicts) :-
	non_user_var(Ground),
	nonvar(Ground),
	functor(Ground,F,A),
	functor(NonGround,F,A),
	ground_to_nonground(A,Ground,NonGround,VarDicts0,VarDicts).

ground_to_nonground(0, _Grond, _NonGround, VarDicts, VarDicts).
ground_to_nonground(N0, Ground, NonGround, VarDicts0, VarDicts) :-
	N0 > 0,
	arg(N0,Ground,GArg),
	arg(N0,NonGround,NArg),
	ground_to_nonground(GArg,NArg,VarDicts0,VarDicts1),
	N is N0 - 1,
	ground_to_nonground(N,Ground,NonGround,VarDicts1,VarDicts).

% nonground_to_ground(+NonGround, -Ground) <-- 
% 	maps a term with logical variables to a term with user-variables
% 	inverse predicate to ground_to_nonground/2

% nonground_to_ground(+NonGround, -Ground, +VarDicts0, -VarDicts) <--
%	VarDicts stores the assignments

nonground_to_ground(NonGround, Ground) :-
	nonground_to_ground(NonGround,Ground,[],_VarDicts).

nonground_to_ground(Var, Ground, VarDicts0, VarDicts) :-  % logical variable maps
	var(Var),                                           % enumerated user-variable
	next_user_var(Ground),                           
	lookup_atom_atom(Ground,Var,VarDicts0,VarDicts),
        !.
nonground_to_ground(Var1, Var2, VarDicts0, VarDicts) :-
        var(Var1),
        var(Var2),
        !,
        lookup_var_var(Var2,Var1,VarDicts0,VarDicts).
nonground_to_ground(NonGround, Ground, VarDicts0, VarDicts) :-
	nonvar(NonGround),
	non_user_var(NonGround),
	functor(NonGround,F,A),
	functor(Ground,F,A),
	nonground_to_ground(A,NonGround,Ground,VarDicts0,VarDicts).

nonground_to_ground(0, _NonGrond, _Ground, VarDicts, VarDicts).
nonground_to_ground(N0, NonGround, Ground, VarDicts0, VarDicts) :-
	N0 > 0,
	arg(N0,NonGround,NArg),
	arg(N0,Ground,GArg),
	nonground_to_ground(NArg,GArg,VarDicts0,VarDicts1),
	N is N0 - 1,
	nonground_to_ground(N,NonGround,Ground,VarDicts1,VarDicts).

:- ground_to_nonground(f(x,f(x,z)),NonGround,[],Di),nonground_to_ground(NonGround,Ground,Di,Di).
:- ground_to_nonground(i(x)*(x*y)=>y,NonGround,[],Di),ground_to_nonground(x*1=>x,NonGround2,Di,Di).
:- nonground_to_ground(f(A,f(A,B)),X).
