
% term.pl
%
% general predicates which handle terms
%
% contains:
%     	is_term/1,
%       is_termlist/1
%       is_ground_term/1
%  	between/3,
%	occurs_in/2,
%     	list_occurs_in/2,
%	subterm/2,
%	prefix/2,
%	substitute/4,
%	perm_congruence/2,
%	renaming/2,
%	lookup_atom_atom/4,
%	lookup_atom_number/4.
%       insertDs/2,
%	memberDs/2										

:- module(term,[is_term/1,
	is_termlist/1,
	is_ground_term/1,
	is_ground_termlist/1,
	between/3,
	occurs_in/2,
     	list_occurs_in/2,
	subterm/2,
	subterm/3,
	substitute_pos/4,
	perm_congruence/2,
	renaming/2,
	lookup_atom_atom/4,
	lookup_var_var/4,
	lookup_atom_number/4,
	lookupGenerate/3,
	lookupGenerate/4,
	insertDs/2,
	memberDs/2]).

:- use_module(library(lists)).

% is_term(Term) <--
%     Term is either a variable, a constant or a function which arguments are terms

is_term(Var) :-
	var(Var).
is_term(Const) :-
	atomic(Const).
is_term(Function) :-
        nonvar(Function),
	functor(Function,_F,A),
	A > 0,
	is_term(1,A,Function).

is_term(A, A, _Function).
is_term(N0, A, Function) :-
	arg(N0,Function,Arg),
	is_term(Arg),
	N is N0 + 1,
	is_term(N,A,Function).

%
% is_termlist/1
%

is_termlist([]).
is_termlist([T|Ts]) :-
    is_term(T),
    is_termlist(Ts).

%
% is_ground_term/1 <--
%

is_ground_term(T) :-
    \+ is_non_ground_term(T).

is_non_ground_term(T) :-
    subterm(T,_Pos,S),
    var(S).

%
%is_ground_termlist/1 <--
%

is_ground_termlist([]).
is_ground_termlist([T|Ts]) :-
    is_ground_term(T),
    is_ground_termlist(Ts).

%
% between(+Low, +High, -X) <-- X is an integer with Low <= X and X <= High

between(Low, High, X) :-
	integer(Low),
	integer(High),
	var(X),
	Low =< High,
	between_1(Low,High,X).
between(Low, High, X) :-
	integer(Low),
	integer(High),
	integer(X),
	Low =< X,
	X =< High.

between_1(Low, _High, Low).
between_1(Low0, High, X) :-
	Low0 < High,
	Low1 is Low0 + 1,
	between_1(Low1,High,X).

:- \+ between(1,3,4).
:- between(1,5,X), X=3.

% occurs_in(+Var, +Term) <--
%	Var is a variable which occurs in Term
%

occurs_in(V, T) :-
	var(T),
	!, 
	T==V.
occurs_in(V, T) :-
	T =.. [_|Arg_Ts],
	occurs_in_list(V,Arg_Ts).

occurs_in_list(V, [T|_Term]) :-
	occurs_in(V,T).
occurs_in_list(V, [_T|Ts]) :-
	occurs_in_list(V,Ts).

:- occurs_in(X,f(X,Y)).

% list_occurs_in(VarLs, Term) <--
%         occurs_in for all variables in VarLs

list_occurs_in([], _Term).
list_occurs_in([Var|VarLs], Term) :-
	var(Var),
	occurs_in(Var,Term),
	list_occurs_in(VarLs,Term).

% subterm(+Term, ?Pos, -Sub) <-- Sub is subterm of Term at position Pos
%

subterm(Term, Sub) :-
	subterm(Term, _Pos, Sub).

subterm(Term, [], Term).
subterm(Term, [P|Pos], Sub) :-
	nonvar(Term),
	functor(Term,_F,A),
	between(1,A,P),
	arg(P,Term,Term_at_P),
	subterm(Term_at_P,Pos,Sub).

:- subterm(g(a,f(c,d),i(e,f(a,a))),Pos,f(a,a)), Pos=[3,2].

% substitute_pos(+Term0, ?Pos, +Subst, -Term) <--
%	Pos is the position where to substitute Subst in Term0

substitute_pos(_Term, [], Subst, Subst).
substitute_pos(Term0, [P|Pos], Subst, Term) :-
	nonvar(Term0),
	functor(Term0,F,A),
	functor(Term,F,A),
	between(1,A,P),
	substitute_pos(A,P,Term0,Pos,Subst,Term).

substitute_pos(0,_P,_Term0,_Pos,_Subst,_Term).
substitute_pos(N0, P, Term0, Pos, Subst, Term) :-
	N0 > 0,
	arg(N0,Term0,TArg0),
	arg(N0,Term,TArg),
	N0=:=P,
	substitute_pos(TArg0,Pos,Subst,TArg),
	N is N0 - 1,
	substitute_pos(N,P,Term0,Pos,Subst,Term).
substitute_pos(N0, P, Term0, Pos, Subst, Term) :-
	N0 > 0,
	arg(N0,Term0,TArg0),
	arg(N0,Term,TArg),
	N0=\=P,
	TArg0=TArg,
	N is N0 - 1,
	substitute_pos(N,P,Term0,Pos,Subst,Term).

:- substitute_pos(g(a,f(c,d),i(e,f(a,a))),[3,2],x,g(a,f(c,d),i(e,x))).
:- substitute_pos(i(i(U))*(i(U)*Y),[1],U,T).

% perm_congruence(+S, +T) <-- T is gained by permutations of subterms of S

perm_congruence(S1, S2) :-		  % check wether corresponding
	var(S1),                     % variables are equal
	var(S2),
	S1==S2.
perm_congruence(S, T) :-
	nonvar(S), 
	nonvar(T),
	S =.. [F|SArgs],
	T =.. [F|TArgs],
	perm_congruence_list(SArgs,TArgs).  % permutations of subterms

perm_congruence_list([], []).
perm_congruence_list(SArgs0, [T|TArgs]) :-
	select(S,SArgs0,SArgs),
	perm_congruence(S,T),
	perm_congruence_list(SArgs,TArgs).

:- perm_congruence(f(f(a,b),g(c,d)),f(g(d,c),f(a,b))).
:- perm_congruence(f(f(x,y),z),f(z,f(y,x))).

% renaming(+X, +Y, ?Var_dicts) <--	

%	Y is a renaming of  X
%	i.e. X has the same structure, but other
%	variables than Y
%	Var_dicts is a variable list, which
%	contains the necessary assignments to
%	provide a correct renaming

renaming(Term1, Term2) :-
	renaming(Term1, Term2, [],_Pairdicts).

renaming(S, T, VarDicts0, VarDicts) :-
	var(S), 
	var(T),
	lookup_atom_atom(S,T,VarDicts0,VarDicts).
renaming(S, T, VarDicts0,VarDicts) :-
	nonvar(S),
	functor(S,F,A),
	functor(T,F,A),
	renaming(A,S,T,VarDicts0,VarDicts).

renaming(0,_S,_T,_VarDicts0,_VarDicts).
renaming(N0, S, T, VarDicts0,VarDicts) :-
	N0 > 0,
	arg(N0,S,SArg),
	arg(N0,T,TArg),
	renaming(SArg,TArg,VarDicts0,VarDicts1),
	N is N0 - 1,
	renaming(N,S,T,VarDicts1,VarDicts).

% lookup_atom_atom(+Key, +Value, +PairDicts, -NewPairDicts) <--
%	PairDicts contains pairs of atoms (i.e variables or constants)
%	NewPairDicts is the  dictionary given after adding Key & Value to PairDicts
%        	Key & Value define a key-value pair 
%        	is this pair new it is added at the end to PairDicts 
%        	FAILs when the specified Key (or Value) exists already  
%        
%	lookup_atom_atom is used to handle assignments of variables, e.g copy_term

lookup_atom_atom(Key, Value, [], [Key-Value]).
lookup_atom_atom(Key, Value, [KeyValue1|PairDicts0], PairDicts) :-
    nonvar(KeyValue1),
    KeyValue1=Key1-Value1,
    Key == Key1,
    !,
    Value == Value1,
    PairDicts = [Key-Value|PairDicts0].
lookup_atom_atom(Key, Value, [KeyValue1|PairDicts0], PairDicts) :-
    nonvar(KeyValue1),
    KeyValue1=Key1-Value1,
    Value == Value1,
    !,
    Key == Key1,
    PairDicts = [Key-Value|PairDicts0].
lookup_atom_atom(Key, Value, [Pair|PairDicts0], [Pair|PairDicts]) :-
    nonvar(Pair),
    lookup_atom_atom(Key,Value,PairDicts0,PairDicts).

% lookup_var_var(?Key, +Value, +PairDicts, -NewPairDicts) <--
%	PairDicts contains pairs of atoms (i.e variables or constants)
%	NewPairDicts is the  dictionary given after adding Key & Value to PairDicts
%        	Key & Value define a key-value pair 
%        	is this pair new it is added at the end to PairDicts 
%        	FAILs when the specified Key (or Value) exists already  
%        
%	lookup_atom_atom is used to handle assignments of variables, e.g copy_term

lookup_var_var(Key, Value, [], [Key-Value]).
lookup_var_var(Key, Value, [KeyValue1|PairDicts0], PairDicts) :-
    nonvar(KeyValue1),
    KeyValue1=Key1-Value1,
    Key == Key1,
    !,
    Value == Value1,
    PairDicts = [Key-Value|PairDicts0].
lookup_var_var(Key, Value, [KeyValue1|PairDicts0], PairDicts) :-
    nonvar(KeyValue1),
    KeyValue1=Key1-Value1,
    Value == Value1,
    !,
    Key = Key1,
    PairDicts = [Key-Value|PairDicts0].
lookup_var_var(Key, Value, [Pair|PairDicts0], [Pair|PairDicts]) :-
    nonvar(Pair),
    lookup_var_var(Key,Value,PairDicts0,PairDicts).

%lookup_var_var(Key, Value, [], [Key-Value]).
%lookup_var_var(Key, Value, [Key1-Value1|PairDicts0], PairDicts) :-
%	Key == Key1,
%	!,
%	Value == Value1,
%	PairDicts = [Key-Value|PairDicts0].
%lookup_var_var(Key, Value, [Key1-Value1|PairDicts0], PairDicts) :-
%	Value == Value1,
%	!,
%	Key == Key1,
%	PairDicts = [Key-Value|PairDicts0].
%lookup_var_var(Key, Value, [Pair|PairDicts0], [Pair|PairDicts]) :-
%	lookup_var_var(Key,Value,PairDicts0,PairDicts).

:- lookup_atom_atom(a,b,[a-b,c-d,e-f,g-1,h-2,i-3,j-4,k-5,l-6,u-7],Dicts).
:- lookup_atom_atom(b,c,[a-b],Xs). 
:- \+ lookup_atom_atom(a,c,[a-b],_).
:- lookup_atom_atom(A,a,[A-a],Xs),lookup_atom_atom(b,B,[A-a],Ys).
:- \+ lookup_atom_atom(B,a,[A-a],Xs).
:- renaming(1*X=X,1*Y=Y).
:- renaming(f(X,Y,Z),f(Y,Z,X)).
:- \+ renaming(f(X,Y,Z),f(Y,Z,Y)).

% lookup_atom_number(+Key, -Number, +PairDicts, -NewPairDicts) <--
% 	PairDicts contains pairs with an atom as key 
% 	and a number as Value
% 	NewPairDicts is the  dictionary given after adding Key & Value
% 	Key is a atom to add to PairDict 
% 	i Key is new, then it is inserted in NewPairDicts at the end, otherwise		
% 	its number value is incremented
% 
% 	lookup_atom_number is used to handle the number of occurences of atoms

lookup_atom_number(Key, 1, [], [Key-1]).              % insert a new atom with # = 1
lookup_atom_number(Key, Number, [Key1-Number1|PairDicts0], PairDicts) :-  
	Key == Key1,                                           % same atom found
	!,
	Number = Number1,
	Number2 is Number +1,
	PairDicts = [Key-Number2|PairDicts0].
lookup_atom_number(Key, Number, [Pair|PairDicts0], [Pair|PairDicts]) :-
	lookup_atom_number(Key,Number,PairDicts0,PairDicts).

:- lookup_atom_number(a,N,[b-2,c-1,C-2,A-4],Xs).
:- lookup_atom_number(C,N,[b-2,c-1,C-2,A-4],Xs).

% lookupGenerate(+VarKey, ?VarValue, +VarDicts0, -VarDicts) <--
% 	VarDicts0 contains pairs with a user-variabe as key and
% 	a 'normal' variable as Value
%	if neither VarKey nor VarValue occure in VarDicts0 these pair is added
%	at the end of VarDicts0 to gain VarDicts
%
% 	lookupGenerate is used in ground_to_nonground/3 and therefor
% 	it is necessary to generate logical variables in this predicate
%

lookupGenerate(VarKey, VarValue, VarDicts) :-
	lookupGenerate(VarKey,VarValue,[],VarDicts).

lookupGenerate(VarKey, VarValue, [], [VarKey-VarValue]).       % insert a new pair
lookupGenerate(VarKey, VarValue, [VarKey1-VarValue1|VarDicts0], VarDicts) :-  % same key
        VarKey==VarKey1,
	!,
	VarValue=VarValue1,
	VarDicts = [VarKey-VarValue|VarDicts0].           % generate the same logical variable
lookupGenerate(VarKey, VarValue, [VarKey1-VarValue1|VarDicts0], VarDicts) :-  % same value
	VarValue1==VarValue,
	!,
	VarKey==VarKey1,
	VarDicts = [VarKey-VarValue|VarDicts0].                                  
lookupGenerate(VarKey, VarValue, [VarPair|VarDicts0], [VarPair|VarDicts]) :-
	lookupGenerate(VarKey,VarValue,VarDicts0,VarDicts).

:- lookupGenerate(b,X,[a-A,b-B],VarDicts), X=B.
:- \+ lookupGenerate(c,B,[a-A,b-B],VarDicts).

% insertDs(X,Xs) <--
% 	Xs is a list with a tail-variable and X is added at the end

insertDs(X,Xs) :-
	var(Xs),
	Xs = [X|_].
insertDs(X,Xs0) :-
	nonvar(Xs0),
	Xs0 = [_|Xs],
	insertDs(X,Xs).

:- insertDs(a,Xs),insertDs(b,Xs).

% memberDs(X,Xs) <--
% 	X is a member of a list with tail-variable Xs

memberDs(X,Xs) :-
	nonvar(Xs),
	Xs = [X|_].
memberDs(X,Xs0) :-
	nonvar(Xs0),
	Xs0 = [_|Xs],
	memberDs(X,Xs).

:- memberDs(a,[a,b,c|Xs]).
:- \+ memberDs(b,[a,c|Xs]).
:- \+ memberDs(A,Xs).


