%----------------------------------------------------------------------------
% read_arm_as_list/2
% read_arm_as_list(+File,-ListOfContentOfFile)
% the file File (which may be either a file name or an already opened stream)
% is read and 'cleaned', that means all comments
% are wiped and the result is given in ListOfContentOfFile.
%----------------------------------------------------------------------------

:- module(tptp_parser,[read_arm_as_list/2,
	mb_parser/2]).

:- assertz(user:file_search_path(basic,'.')).

:- use_module(basic(general)).
	
read_arm_as_list(F,X) :-
        atom(F),
        open(F,read,S),
        write('[reading '), write(F),
	read_arm_as_list(S,X),
        write(']'), nl,
        close(S),!.
read_arm_as_list(S,X) :-
        getclean(S,C),
        read_arm_as_list1(S,C,X),
	!.

read_arm_as_list1(_,-1,[]).
read_arm_as_list1(S,C,[C|T]) :-
        getclean(S,C1),
        read_arm_as_list1(S,C1,T).

getclean(S,C) :-
        get0(S,C1),
        (C1 = -1 -> C = C1
        ;C1 = 37 -> getcomment(S,C)
        ;C1 < 33 -> C = 32
        ;           C = C1
        ).

getcomment(S,C) :-
	get0(S,C1),
	( C1 = -1         -> C = C1
	; newlinechar(C1) -> C = 32
        ;                    getcomment(S,C)
        ).

newlinechar(10).
newlinechar(13).

%%%%%%%%%%%%%
%
% mb_parser <---

mb_parser(List,State) :-
	parse_mb(State,List,[]).
	

%%%%%%%%%%%%%%%%%%%%
%
% parse_mb <--
%


parse_mb(eval(Atoms,Equations,HUniverse)) -->
	dfg_spaces(_),
	parse_clauses(Atoms,Equations,HUniverse),
	dfg_spaces(_).

parse_clauses(Ats,Eqs,HUs) -->
	parse_clause(Ats,Eqs,HUs),
	".".
parse_clauses(Ats,Eqs,HUs) -->
	parse_clause(At0s,Eq0s,HU0s),
	".",
	dfg_spaces(_),
	{ ! },
	parse_clauses(At1s,Eq1s,HU1s),
	{ merge(At0s,At1s,Ats) },
	{ merge(Eq0s,Eq1s,Eqs) },
	{ merge(HU0s,HU1s,HUs) }.

dfg_spaces(_) -->
	" ",
	dfg_spaces(_).
dfg_spaces(_) -->
	"",
	{ ! }.

parse_clause(Ats,Eqs,HUs) -->
	"[",
	parse_literals(Ats,Eqs,HUs),
	"]".

parse_literals(At,Eqs,HUs) -->
	parse_literal(At,Eqs,HUs).
parse_literals(Ats,Eqs,HUs) -->
	parse_literal(At0s,Eq0s,HU0s),
	",",
	parse_literals(At1s,Eq1s,HU1s),
	{ merge(At0s,At1s,Ats) },
	{ merge(Eq0s,Eq1s,Eqs) },
	{ merge(HU0s,HU1s,HUs) }.

parse_literal(Ats,Eqs,HUs) -->
	tptp_sign,
	parse_atom(Ats,Eqs,HUs).

parse_atom([],[Eq],HUs) -->
	dfg_equation(Eq,HUs).
parse_atom([At],[],HUs) -->
	dfg_pred_expr(At,HUs).
parse_atom([X],[],[]) -->
	dfg_zero_op(X).

tptp_sign -->
	"++".
tptp_sign -->
	"--".

%%%%%%%%%%%%%%%%%%%%%%
%
% predicates from dfg_parser.pl
%

dfg_zero_op(true) -->
	"true".
dfg_zero_op(false)  -->
	"false".
dfg_zero_op(X) -->
	dfg_op_sym(X).

dfg_op_sym(X) -->
	dfg_identifier(X).

dfg_pred_expr(At,HUs) -->
	dfg_identifier(Pred),
	"(",
	dfg_termlist(TermList,HUs,_A),
	")",
	{ At =.. [Pred|TermList] }.
dfg_pred_expr(At,[]) -->
        dfg_identifier(Pred),
	{ At =.. [Pred] }.

dfg_equation(L=R,HUs) -->
	"equal(",
	dfg_term(L,HU0s),
	",",
	dfg_term(R,HU1s),
	")",
	{ merge(HU0s,HU1s,HUs) }.

dfg_termlist([H],HUs,1) -->
	dfg_term(H,HUs).
dfg_termlist([H|T],HUs,N) -->
	dfg_term(H,HU0s),
	",",
	dfg_termlist(T,HU1s,N0),
	{ merge(HU0s,HU1s,HUs), N is N0 + 1 }.

dfg_term(X,HUs)  -->
	dfg_fun_sym(Funct),
	"(",
	dfg_termlist(Args,HU1s,A),
	")",
	{ X =.. [Funct|Args]},
	{ merge([Funct/A],HU1s,HUs) }.
dfg_term(Funct,[Funct/0])  -->
	dfg_constant(Funct).

dfg_constant(X) -->
	dfg_identifier(X).

dfg_fun_sym(X) -->
	dfg_identifier(X).

dfg_identifierlist([H|T]) -->
	dfg_identifier(H),
	(",", dfg_identifierlist(T) ; { T = [] }),
	{ ! }.

dfg_identifier(X) -->
	dfg_identifier_a(H),
	dfg_identifier(T),
	{ name(H,[HN]), name(T,TN), name(X,[HN|TN]), ! }.
dfg_identifier(X) -->
	dfg_identifier_a(X).

dfg_identifier_a(X) -->
	dfg_letter(X),
	{ ! }.
dfg_identifier_a(X) -->
	dfg_digit(N),
	{N1 is N+"0", name(X,[N1]), ! }.
dfg_identifier_a(X) -->
	dfg_special_symbol(X).

dfg_letter(X) -->
	dfg_var(X).
dfg_letter(X) -->
	dfg_const(X).

dfg_var(X) -->
	[L],
	{ "A"=<L, L=<"Z", name(X,[L]) }.
dfg_const(X) -->
	[L],
	{ "a"=<L, L=<"z", name(X,[L]) }.

dfg_number(C) -->
	dfg_digit(H),
	dfg_number(T),
	{exb(T,F), C is H*F+T}.
dfg_number(C) -->
	dfg_digit(C).

dfg_digit(X) -->
	[C],
	{ "0"=<C, C=<"9", X is C - "0" }.

dfg_special_symbol('_') -->
	"_".

exb(N,10) :- N < 10, !.
exb(N, E) :- N1 is N // 10, exb(N1,E1), E is E1*10, !.

:- phrase(dfg_termlist(X,HU,A),"a,b").
:- phrase(parse_literals(A,E,H),"++love(f(a)),++hate(f(b))").
%:- read_arm_as_list('~/FWFmodel/examples/perm1.arm',X), mb_parser(X,S).
