============================================================================
%  dfg_parser.pl
%  definite clause grammar parser for the
%  "Common Syntax of the DFG-Schwerpunktprogramm `Deduktion'"
%
%  Written by Michael Christen, October 1996.
%  e-mail: christen@mpi-sb.mpg.de, www: http://www.mpi-sb.mpg.de/~christen/
%============================================================================


% missing features for this version:
%-----------------------------------
% - no input statement
% - no comments of the form /* */
% - description part is optional (which should not be optional)


:- module(dfg_parser, [
        read_file_as_list/2,
	parse_dfg/2,
	parse_dfg_tail/2,
	saturate_info/8
   ]).

/*
Documentation:
--------------

Parsing a DFG-file consist of two steps:
1st : reading the entire file and creating a list with its content
2nd : parsing the list and creating a term with information of symbols and clauses

to 1)
To read an input-file we use the predicate read_file_as_list/2
If you already opened a stream, you may read the last part of the file also
with read_file_as_list/2, but doing so, you must close the stream yourself.

to 2)
Some parsers may make use of the prolog read-predicate "read" to scan an input file.
If the type of the input file is not known at this moment, the first term of a dfg-file
read by a parser is the statement "begin_problem(label).".
If this happens you can read the last part of the file and parse it with parse_dfg_tail/2.
Otherwise (if you read the entire file including the statement begin_problem(_))
you must parse it with parse_dfg/2.


Suppose you want to parse a file 'test.dfg'. You can do so by using the following predicates:

        read_file_as_list('test.dfg',FileContent),   % read file and get result as list
	parse_dfg(FileContent,X),                    % parse content with DCG's

The result X  is a term and has the following form:

X = problem(
       NickName,                                                 % a symbol
       description(Name,Author,Version,Logic,Status,Descr,Date), % texts as symbols
       logical_part(
          symbol_list(
             Functions,   % = [ Symbol/Arity | .. ]
	     Predicates,  % = [ Symbol/Arity | .. ]
	     Sorts,       % = [ Symbol/Arity | .. ]
	     Operators,   % = [ Symbol/Arity | .. ]
	     Quantifiers  % = [ Symbol/Arity | .. ]
	  ),
          Declarations,   %
          Formulae,       % = [ formulae(Origin , 'pl1'      , Body)    | .. ]
          Clauses,        % = [ formulae(Origin , ClauseType , Clauses) | .. ]
          Proofs          % = [ formulae('proof', Type       , Proof)   | .. ]
       ),
       Settings           % = [settings(Label,Text)|..]
   ).

Example: to get the lists of 
 - P: Predicates
 - F: Functions
 - C: Clauses
with a problem descrition X, you can instantiate P,F,C with:

X = problem(_,_,logical_part(symbol_list(F,P,_,_,_),_,_,C,_),_).

see the predicate saturate_info/8 as an example how to process
information given by parse_dfg/2.



Term translation:
-----------------

Terms read by parse_dfg have the same form as described at the DFG-paper
except quantifiers and the 'and' and 'or' - junctions:

Quantifier (Q='forall' or Q='exists') - terms Q([X1,...,Xn],Term)
are translated to Q(X1,Q(X2, ... Q(Xn,Term) .. )) and
the n-ary junctions 'and' and 'or' are translated to binary junctions, for
example: and(p1,p2,p3,p4) will be translated to and(p1,and(p2,and(p3,4))).




Installation:
-------------


If you want to use this parser for SATURATE you only have to paste the following
goal, clause and predicate into the file 'input.pl':


:- use_module(dfg_parser2, [ % <DFG>
        read_file_as_list/2,
	parse_dfg_tail/2,
	saturate_info/8
   ]).


% this clause must be #3 (occuring earlier than the general case of parse/4)
parse(Str, _F, _M, begin_problem(_ProblemLabel) ):-                 % <DFG: parse dfg-files>
        read_file_as_list(Str,FileTailList),                        % read file
	close(Str),                                                 % close stream
	parse_dfg_tail(FileTailList,ParsedFile),                    % parse content with DCG's
	saturate_info(ParsedFile,_Pred,_Func,Ax,Hyp,Goals,ThF,GoF), % saturate relevant information
	((ThF = [], GoF = []) ->                                    % if no formulae
	    treat_input_terms(Ax,theory),                           % treat axioms
	    treat_input_terms(Hyp,theory),
	    treat_input_terms(Goals,(goal))
	;
	    treat_input_terms(ThF, theory),                         % but prefer formulae if existent
	    treat_input_terms(GoF, (goal))
	).


treat_input_terms([Term|Tail],Type) :- % <DFG: treat a list of clauses>
	treat_input(Term, _, Type),
	treat_input_terms(Tail,Type).
treat_input_terms([],_).

*/




:- use_module(library(lists), [member/2]).
:- use_module(library(lists),   [select/3]).


%----------------------------------------------------------------------------
% read_file_as_list/2
% read_file_as_list(+File,-ListOfContentOfFile)
% the file File (which may be either a file name or an already opened stream)
% is read and 'cleaned', that means all comments
% are wiped and the result is given in ListOfContentOfFile.
%----------------------------------------------------------------------------

read_file_as_list(F,X) :-
        atom(F),
        open(F,read,S),
        write('[reading '), write(F),
	read_file_as_list(S,X),
        write(']'), nl,
        close(S),!.
read_file_as_list(S,X) :-
        getclean(S,C),
        read_file_as_list1(S,C,X),
	!.

read_file_as_list1(_,-1,[]).
read_file_as_list1(S,C,[C|T]) :-
        getclean(S,C1),
        read_file_as_list1(S,C1,T).

getclean(S,C) :-
        get0(S,C1),
        (C1 = -1 -> C = C1
        ;C1 = 37 -> getcomment(S,C)
        ;C1 < 33 -> C = 32
        ;           C = C1
        ).

getcomment(S,C) :-
	get0(S,C1),
	( C1 = -1         -> C = C1
	; newlinechar(C1) -> C = 32
        ;                    getcomment(S,C)
        ).

newlinechar(10).
newlinechar(13).



%----------------------------------------------------------------------------
% parse_dfg(+List,-Term)
% parse_dfg_tail(+List,-Term)
% parse a list of characters
% and create a term constructed from parsed information
%----------------------------------------------------------------------------


parse_dfg(List,Clauses) :-
	dfg_problem(Clauses,List,[]).

parse_dfg_tail(List,Clauses) :-
	dfg_problem_tail(Clauses,List,[]).



%----------------------------------------------------------------------------
% saturate_info/8
% saturate_info(+ParseResult,-Pred,-Func,-Ax,-Hyp,-Conj,-TheoryFormula,-GoalFormula)
% create Saturate relevant information
%----------------------------------------------------------------------------

saturate_info(X,Predicates,Functions,Ax,Hyp,Conj,TheoryForm,GoalForm) :-
	saturate_substitution(Subst),
	X = problem(
	       _NickName,
	       description(_Name,_Author,_Version,_Logic,_Status,_Descr,_Date),
	       logical_part(
	          symbol_list(Functions,Predicates,_Sorts,_Operators,_Quantifiers),
		  _Declarations,
		  Formulae,
		  Clauses,
		  _Proofs
	       ),
	       Settings
	   ),
	(member(formulae(axioms,cnf,AxCl),Clauses) -> true ; AxCl = []),
	(member(formulae(conjectures,cnf,ConjCl),Clauses) -> true ; ConjCl = []),
	(member(formulae(axioms,pl1,AxForm),Formulae) -> true ; AxForm = []),
	(member(formulae(conjectures,pl1,ConjForm),Formulae) -> true ; ConjForm = []),
	(member(settings(dfg,Hypolist),Settings) -> true ; Hypolist =[[]]),
	append2(Hypolist,Hypolabels),
	!,
	% don't mix formulas with clauses, bring both into a form acceptable for saturate.
	%
	% Clauses:
	% divide ConjCl into hypothesis and rest with Hypolabels
	split_formulae(ConjCl,Hypolabels,HCl,RCl),
	% strip off labels:
	stripoff_labels(AxCl, AxR),
	stripoff_labels(HCl , HypR),
	stripoff_labels(RCl , ConjR),
	substitute(AxR  , Subst, Ax),
	substitute(HypR , Subst, Hyp),
	substitute(ConjR, Subst, Conj),
	%
	% Formulas:
	(ConjForm = [] ->
	    GoalForm = [],
	    (AxForm = [] ->
		TheoryForm = []
	    ;
	        stripoff_labels(AxForm,Antecedent),
		substitute( Antecedent, Subst, TheoryForm)
	    )
	;
	    stripoff_labels(ConjForm,ConjFormNL),
	    make_bin_junction(ConjFormNL,and,Succedent),
	    (Succedent = not(NotSuccedent) ->
		substitute( [NotSuccedent], Subst, GoalForm)
	    ;
	        substitute( [not(Succedent)], Subst, GoalForm)
	    ),
	    (AxForm = [] ->
		TheoryForm = []
	    ;
	        stripoff_labels(AxForm,Antecedent),
		substitute( Antecedent, Subst, TheoryForm)
	    )
	).

saturate_info(_,[],[],[],[],[],[],[]).

saturate_substitution([(not,~),(forall,all),(implies,=>),(implied,<=),(equiv,==),(equal,=)]).

stripoff_labels([],[]).
stripoff_labels([_:A|T],[A|TS]) :-
	stripoff_labels(T,TS).

split_formulae([],_,[],[]).
split_formulae([L:F|T],Labels,[L:F|I],O) :-
	member(L,Labels),
	split_formulae(T,Labels,I,O).
split_formulae([H|T],Labels,I,[H|O]) :-
	split_formulae(T,Labels,I,O).




%----------------------------------------------------------------------------
%
% DFG - DCG
%
%----------------------------------------------------------------------------


dfg_problem(problem(Label, Description, Logic, Settings)) -->
	dfg_spaces(_),
	"begin_problem(",
	dfg_spaces(_),
	dfg_label(Label),
	dfg_spaces(_),
	")",
	dfg_spaces(_),
	".",
	{ ! },
	dfg_problem_tail(problem(_, Description, Logic, Settings)).

dfg_problem_tail(problem(_, Description, Logic, Settings)) -->
	dfg_spaces(_),
	dfg_description(Description), % description part may contain spaces
	dfg_resttext(Rest),           % special handling of part of text without spaces
	"end_problem.",
	dfg_spaces(_),
	{ wipespace(Rest,Clean),
	  !,
	  dfg_problem_tail2((Logic,Settings),Clean,[])
        }.

dfg_problem_tail2((Logic, Settings)) -->
	dfg_logical_part(Logic),
	dfg_settings_list(Settings),
	{ ! }.

dfg_description(X)  -->
	"list_of_descriptions.",
	dfg_spaces(_),
	 "name(",        dfg_text(NameN),       ").",                     { name(Name,    NameN)    },
	dfg_spaces(_),
	 "author(",      dfg_text(AuthorN),     ").",                     { name(Author,  AuthorN)  },
	dfg_spaces(_),
	("version(",     dfg_text(VersionN),    ")." ; { Version = [] }), { name(Version, VersionN) },
	dfg_spaces(_),
	("logic(",       dfg_text(LogicN),      ")." ; { Logic   = [] }), { name(Logic,   LogicN)   },
	dfg_spaces(_),
	 "status(",      dfg_log_state(Status), ").",
	dfg_spaces(_),
	 "description(", dfg_text(DescrN),      ").",                     { name(Descr,   DescrN)   },
	dfg_spaces(_),
	("date(",        dfg_text(DateN),       ")." ; { Date    = [] }), { name(Date,    DateN)    },
	dfg_spaces(_),
	"end_of_list.",
	{ X = description(Name,Author,Version,Logic,Status,Descr,Date), !}.
dfg_description(description(unknown,unknown,unknown,unknown,unknown,unknown,unknown))  --> { ! }.

dfg_log_state(X) -->
	{( X = satisfiable ; X = unsatisfiable ; X = unknown), name(X,N)},
	N,
	{ ! }.

dfg_settings_list([H|T]) -->
	dfg_settings(H),
	dfg_settings_list(T),
	{ ! }.
dfg_settings_list([]) -->
	{ ! }.

dfg_settings(settings(dfg,X)) -->
	"list_of_general_settings.",
	dfg_setting_entrylist(X),
	"end_of_list.",
	{ ! }.
dfg_settings(settings(Label,Text)) -->
	"list_of_settings",
	"(",
	dfg_setting_label(Label),
	").",
	dfg_text(Text),
	"end_of_list.".

dfg_setting_entrylist([H|T]) -->
	dfg_setting_entry(H),
	(dfg_setting_entrylist(T) ; { T = [] }),
	{ ! }.

dfg_setting_entry(X) -->
	"hypothesis[",
	dfg_labellist(X),
	"].",
	{ ! }.
dfg_setting_entry([]) -->
	"hypothesis[].",
	{ ! }. %% NOT CONSISTENT WITH LANGUAGE DESCRIPTION! (but nessecary)


dfg_labellist([H|T]) -->
	dfg_label(H),
	(",", dfg_labellist(T) ; { T = [] }),
	{ ! }.

dfg_setting_label(Label) -->
	{ Name = "KIV" ; Name = "LEM" ; Name = "PROTEIN" ; Name = "SATURATE" ;
	  Name = "3TAP" ; Name = "SETHEO" ; Name = "SPASS" },
	Name,
	{ name(Label,Name), ! }.

dfg_logical_part(X) -->
	dfg_opt_symbol_list(S),
	(dfg_declaration_list(D) ; { D = [] }),
	dfg_formula_list_list(F),
	dfg_clause_list_list(C),
	dfg_proof_list_list(P),
	{ append_formulae(F,FA),
	  append_formulae(C,CA),
	  append_formulae(P,PA),
          X = logical_part(S,D,FA,CA,PA), ! }.

dfg_formula_list_list(X) -->
	"list_of_formulae(",
	dfg_origin_type(Origin),
	").",
	dfg_formula_list_body(Body),
	"end_of_list.",
	dfg_formula_list_list(T),
	{ X = [formulae(Origin,pl1,Body)|T], ! }.
dfg_formula_list_list([]) -->
	{ ! }.

dfg_clause_list_list(X) -->
	"list_of_clauses(",
	dfg_origin_type(Origin),
	",",
	dfg_clause_type(CType),
	").",
	dfg_clause_list_body(Clauses),
	"end_of_list.",
	dfg_clause_list_list(T),
	{ X = [formulae(Origin,CType,Clauses)|T], ! }.
dfg_clause_list_list([]) -->
	{ ! }.

dfg_proof_list_list(X) -->
	"list_of_proof(",
	dfg_proof_type(Type),
	").",
	dfg_proof_list_body(Proof),
	"end_of_list.",
	dfg_proof_list_list(T),
	{ X = [formulae(proof,Type,Proof)|T], ! }.
dfg_proof_list_list([]) -->
	{ ! }.

dfg_proof_type('resolution') -->
	"resolution".

dfg_justif(X) -->
	dfg_label(X).

dfg_label_list(X) -->
	"[",
	dfg_labellist(X),
	"]".

dfg_opt_symbol_list(X) -->
	"list_of_symbols.",
	("functions[", dfg_functions_list(F), "]." ; { F = [] } ),
	("predicates[", dfg_predicates_list(P), "]." ; { P = [] } ),
	("sorts[", dfg_sorts_list(S), "]." ; { S = [] } ),
	("operators[", dfg_operators_list(O), "]." ; { O = [] } ),
	("quantifiers[", dfg_quantifiers_list(Q), "]." ; { Q = [] } ),
	"end_of_list.",
	{ X = symbol_list(F,P,S,O,Q), ! }.
dfg_opt_symbol_list(X) -->
	{ X = symbol_list([],[],[],[],[]), ! }.

dfg_functions_list(X) -->
	( "(", dfg_fun_sym(S), ",", dfg_arity(A), ")" ; dfg_fun_sym(S), { A = -1 }),
	(",", dfg_functions_list(T) ; { T = [] }),
	{ X = [S/A | T] ,! }.

dfg_predicates_list(X) -->
	( "(", dfg_pred_sym(S), ",", dfg_arity(A), ")" ; dfg_pred_sym(S), { A = -1 }),
	(",", dfg_predicates_list(T) ; { T = [] }),
	{ X = [S/A | T] ,! }.

dfg_sorts_list(X) -->
	( "(", dfg_sort_sym(S), ",", dfg_arity(A), ")" ; dfg_sort_sym(S), { A = -1 }),
	(",", dfg_sorts_list(T) ; { T = [] }),
	{ X = [S/A | T] ,! }.

dfg_operators_list(X) -->
	( "(", dfg_op_sym(S), ",", dfg_arity(A), ")" ; dfg_op_sym(S), { A = -1 }),
	(",", dfg_operators_list(T) ; { T = [] }),
	{ X = [S/A | T] ,! }.

dfg_quantifiers_list(X) -->
	( "(", dfg_quant_sym(S), ",", dfg_arity(A), ")" ; dfg_quant_sym(S), { A = -1 }),
	(",", dfg_quantifiers_list(T) ; { T = [] }),
	{ X = [S/A | T] ,! }.

dfg_arity(X) -->
	"-1",
	{ X = -1, ! }.
dfg_arity(X) -->
	dfg_number(X).

dfg_declaration_list(X) -->
	"list_of_declarations.",
	dfg_declarationlist(X),
	"end_of_list.".

dfg_declarationlist([H|T]) -->
	dfg_declaration(H),
	(",", dfg_declarationlist(T) ; { T = [] }),
	{ ! }.

dfg_declaration(X)  -->
	dfg_subsort_decl(X).
dfg_declaration(X)  -->
	dfg_term_decl(X).
dfg_declaration(X)  -->
	dfg_pred_decl(X).

dfg_subsort_decl(subsort(S1,S2)) -->
	"subsort(",
	dfg_sort_sym(S1),
	",",
	dfg_sort_sym(S2),
	").",
	{ ! }.

dfg_sort_sym(X) -->
	dfg_identifier(X).

dfg_term_decl(X) -->
	"forall(",
	dfg_var_list(Vars),
	",",
	dfg_term_expr(Body),
	")",
	{ create_vars(Vars,VarTup,NewVars),
	  substitute(Body,VarTup,InstBody),
	  make_single_quantor(NewVars,forall,InstBody,X), ! }.
dfg_term_decl(TermExpr) -->
	dfg_term_expr(TermExpr), ".",
	{ ! }.

dfg_term_expr(X) -->
	dfg_sort_sym(Sort),
	"(",
	dfg_term(Term),
	")",
	{ X =.. [Sort,Term], ! }.

dfg_pred_decl(predicate(Pred,Sym)) -->
	"predicate(",
	dfg_pred_sym(Pred),
	",",
	dfg_sort_symlist(Sym),
	").",
	{ ! }.

dfg_sort_symlist([H|T]) -->
	dfg_sort_sym(H),
	(",", dfg_sort_symlist(T) ; { T = [] }),
	{ ! }.

dfg_formula_list_body(X) -->
	"formula(",
	dfg_formula(Form),
	( ",", dfg_formula_label(Label) ; { Label = noname }),
	").",
	dfg_formula_list_body(T), % no comma here!
	{ X = [':'(Label,Form)|T], ! }.
dfg_formula_list_body([]) -->
	{ ! }.

dfg_formula_label(X) -->
	dfg_pred_expr(X).

dfg_label(X) -->
	dfg_pred_expr(X).

dfg_formula(X) -->
	dfg_quant_formula(X).
dfg_formula(X) -->
	dfg_log_op_formula(X).
dfg_formula(X) -->
	dfg_atom(X).

dfg_clause_list_body(X) -->
	"clause(",
	( dfg_cnf_clause(NF) ; dfg_dnf_clause(NF) ),
	( ",", dfg_label(Label) ; { Label = noname } ),
	").",
	dfg_clause_list_body(T),
	{ X = [':'(Label,NF)|T], ! }.
dfg_clause_list_body([]) -->
	{ ! }.

dfg_clause_type(cnf) -->
	"cnf".
dfg_clause_type(dnf) -->
	"dnf".

dfg_cnf_clause(X) -->
	"forall(",
	dfg_var_list(Vars),
	",",
	dfg_cnf_clause_body(Body),
	")",
	{ create_vars(Vars,VarTup,NewVars),
	  substitute(Body,VarTup,InstBody),
	  make_single_quantor(NewVars,forall,InstBody,X), ! }.
dfg_cnf_clause(Body) -->
	dfg_cnf_clause_body(Body),
	{ ! }.

dfg_dnf_clause(X) -->
	"exists(",
	dfg_var_list(Vars),
	",",
	dfg_dnf_clause_body(Body),
	")",
	{ create_vars(Vars,VarTup,NewVars),
	  substitute(Body,VarTup,InstBody),
	  make_single_quantor(NewVars,exists,InstBody,X), ! }.
dfg_dnf_clause(Body) -->
	dfg_dnf_clause_body(Body),
	{ ! }.

dfg_cnf_clause_body(X) -->
	"or(",
	dfg_literallist(LList),
	")",
	{ make_bin_junction(LList,or,X), ! }.
dfg_dnf_clause_body(X) -->
	"and(",
	dfg_literallist(LList),
	")",
	{ make_bin_junction(LList,and,X), ! }.

dfg_literallist([H|T])  -->
	dfg_literal(H),
	(",", dfg_literallist(T) ; { T = [] }),
	{ ! }.

dfg_literal(X) -->
	dfg_un_op(Op),
	"(",
	dfg_atom(Atom),
	")",
	{ X =.. [Op,Atom], ! }.
dfg_literal(X) -->
	dfg_atom(X).

dfg_quant_formula(X) -->
	dfg_quantifier(QuantSym),
	"(",
	dfg_term_list(Terms),
	",",
	dfg_formula(Formula),
	")",
	{ create_vars(Terms,VarTup,NewVars),
	  substitute(Formula,VarTup,InstFormula),
	  make_single_quantor(NewVars,QuantSym,InstFormula,X), ! }.

dfg_quantifier(forall) -->
	"forall".
dfg_quantifier(exists) -->
	"exists".
dfg_quantifier(X) -->
	dfg_quant_sym(X).

dfg_quant_sym(X) -->
	dfg_identifier(X).

dfg_origin_type(axioms) -->
	"axioms".

dfg_origin_type(conjectures) -->
	"conjectures".

dfg_var_list(X) -->
	"[",
	dfg_variable_exprlist(X),
	"]",
	{ ! }.
dfg_variable_exprlist([H|T]) -->
	dfg_variable_expr(H),
	(",", dfg_variable_exprlist(T) ; { T = [] }),
	{ ! }.

dfg_term_list(X) -->
	"[",
	dfg_termlist(X),
	"]",
	{ ! }.

dfg_variable_expr(X) -->
	dfg_sort_sym(SortSym),
	"(",
	dfg_variable_sym(V),
	")",
	{ X =.. [SortSym, V], ! }.
dfg_variable_expr(X) -->
	dfg_variable_sym(X).

dfg_variable_sym(X) -->
	dfg_identifier(X).

dfg_log_op_formula(X) -->
	dfg_un_op(Op),
	"(",
	dfg_formula(Form),
	")",
	{ X =.. [Op,Form], ! }.
dfg_log_op_formula(X) -->
	dfg_bin_op(Op),
	"(",
	dfg_formula(Form1),
	",",
	dfg_formula(Form2),
	")",
	{ X =.. [Op,Form1,Form2], ! }.
dfg_log_op_formula(X) -->
	dfg_n_op(Op),
	"(",
	dfg_formulalist(FormList),
	")",
	{((Op = and ; Op = or) ->
	      make_bin_junction(FormList,Op,X)
	  ;
	      X =.. [Op|FormList]
	)}.

dfg_formulalist([H|T])  -->
	dfg_formula(H),
	(",", dfg_formulalist(T) ; { T = [] }),
	{ ! }.

dfg_un_op(not) -->
	"not".

dfg_bin_op(implies) -->
	"implies".
dfg_bin_op(implied) -->
	"implied".
dfg_bin_op(equiv) -->
	"equiv".

dfg_n_op(and) -->
	"and".
dfg_n_op(or) -->
	"or".
dfg_n_op(X)  -->
	dfg_op_sym(X).

dfg_atom(X)  -->
	dfg_equation(X).
dfg_atom(X)  -->
	dfg_pred_expr(X).
dfg_atom(X)  -->
	dfg_zero_op(X).

dfg_zero_op(true) -->
	"true".
dfg_zero_op(false)  -->
	"false".
dfg_zero_op(X) -->
	dfg_op_sym(X).

dfg_op_sym(X) -->
	dfg_identifier(X).

dfg_pred_expr(X) -->
	dfg_pred_sym(Pred),
	"(",
	dfg_termlist(TermList),
	")",
	{ X =.. [Pred|TermList], ! }.
dfg_pred_expr(X) -->
	dfg_sort_sym(SortSym),
	"(",
	dfg_term(Term),
	")",
	{ X =.. [SortSym,Term], ! }.
dfg_pred_expr(X) -->
	dfg_pred_sym(Pred),
	{ X =.. [Pred] }.

dfg_equation(equal(L,R)) -->
	"equal(",
	dfg_term(L),
	",",
	dfg_term(R),
	")",
	{ ! }.
dfg_pred_sym(X) -->
	dfg_identifier(X).

dfg_termlist([H|T]) -->
	dfg_term(H),
	(",", dfg_termlist(T) ; {T = []}),
	{ ! }.

dfg_term(X)  -->
	dfg_fun_sym(Funct),
	"(",
	dfg_termlist(Args),
	")",
	{ X =.. [Funct|Args], ! }.
dfg_term(X)  -->
	dfg_constant(X).

dfg_constant(X) -->
	dfg_identifier(X).
dfg_fun_sym(X) -->
	dfg_identifier(X).

dfg_identifierlist([H|T]) -->
	dfg_identifier(H),
	(",", dfg_identifierlist(T) ; { T = [] }),
	{ ! }.

dfg_identifier(X) -->
	dfg_identifier_a(H),
	dfg_identifier(T),
	{ name(H,[HN]), name(T,TN), name(X,[HN|TN]), ! }.
dfg_identifier(X) -->
	dfg_identifier_a(X).

dfg_identifier_a(X) -->
	dfg_letter(X),
	{ ! }.
dfg_identifier_a(X) -->
	dfg_digit(N),
	{N1 is N+"0", name(X,[N1]), ! }.
dfg_identifier_a(X) -->
	dfg_special_symbol(X).

dfg_letter(X) -->
	[L],
	{ (("A"=<L, L=<"Z") ; ("a"=<L, L=<"z")), name(X,[L]), ! }.

dfg_number(C) -->
	dfg_digit(H),
	dfg_number(T),
	{exb(T,F), C is H*F+T}.
dfg_number(C) -->
	dfg_digit(C).

dfg_digit(X) -->
	[C],
	{ "0"=<C, C=<"9", X is C - "0" }.

dfg_special_symbol('_') -->
	"_".

exb(N,10) :- N < 10, !.
exb(N, E) :- N1 is N // 10, exb(N1,E1), E is E1*10, !.


dfg_spaces(_) -->
	" ",
	dfg_spaces(_).
dfg_spaces(_) -->
	"",
	{ ! }.

dfg_resttext([H|T]) -->
	[H],
	dfg_resttext(T).
dfg_resttext([]) -->
	"".

dfg_text([]) -->
	"".
dfg_text([H|T]) -->
	[H],
	dfg_text(T).
% dfg_text and dfg_resttext must be fully backtrackable


%----------------------------------------------------------------------------
% helper predicates for DCG's
%----------------------------------------------------------------------------


wipespace([],[]).
wipespace([H|T],W) :-
	H < 32,
	wipespace([32|T],W).
wipespace([32|T],W) :-
	wipespace(T,W).
wipespace([H|T],[H|W]) :-
	wipespace(T,W).

create_vars([],[],[]).
create_vars([H|T],[(H,NewVar)|VT],[NewVar|RestVar]) :-
	create_vars(T,VT,RestVar), !.

substitute(Formula,_,Formula) :-
	var(Formula).
substitute(Formula,VarTup,InstFormula) :-
	atom(Formula),
	member((Formula,InstFormula),VarTup),!.
substitute(Formula,_,Formula) :-
	atom(Formula).
substitute(Formula,VarTup,InstFormula) :-
	functor(Formula,Name,Arity),
	(member((Name,NewName),VarTup) -> true ; NewName = Name),
	functor(InstFormula,NewName,Arity),
	substitute(Formula,Arity,VarTup,InstFormula).

substitute(_,0,_,_).
substitute(F,A,VT,F1) :-
	arg(A,F,Arg),
	substitute(Arg,VT,Arg1),
	arg(A,F1,Arg1),
	A1 is A - 1,
	substitute(F,A1,VT,F1).

split_AS([],[],[]).
split_AS([not(A)|T],[A|AT],ST) :-
	split_AS(T,AT,ST).
split_AS([S|T],AT,[S|ST]) :-
	split_AS(T,AT,ST).

append_formulae([],[]).
append_formulae([formulae(A,B,C)|T],X) :-
	select(formulae(A,B,C1),T,T1),
	append(C,C1,CA),
	!,
	append_formulae([formulae(A,B,CA)|T1],X).
append_formulae([H|T],[H|T1]) :-
	append_formulae(T,T1).


make_bin_junction([],_,true) :- !, fail.
make_bin_junction([Form],_,Form).
make_bin_junction([H1,H2],Op,Form) :-
	Form =.. [Op,H1,H2].
make_bin_junction([H|T],Op,Form) :-
	make_bin_junction(T,Op,Form1),
	Form =.. [Op,H,Form1].

%make_single_quantor(Vars,Quantor,Formula,Result)
make_single_quantor([],_,F,F).
make_single_quantor([H|T],Q,F,Form) :-
	make_single_quantor(T,Q,F,Next),
	Form =.. [Q,H,Next].

append2([],[]).
append2([H|T],R) :-
	append2(T,TR),
	append(H,TR,R).


% EOF
